import os
from collections import defaultdict

from typing import List, Dict
from pathlib import Path

LOCK_HEADER = "# LOCK-BEGIN["
LOCK_FOOTER = "# LOCK-END"
LOCK_WARNING = "]: DON'T MODIFY"


class ImportCollector:
    def __init__(self):
        self._imports = set()
        self._from_imports = defaultdict(set)

    def __bool__(self):
        return bool(self._imports) or bool(self._from_imports)

    def add_import(self, import_clause, as_clause=None):
        if as_clause is not None:
            import_clause += " as " + as_clause
        self._imports.add(import_clause)

    def add_from_import(self, from_clause, import_clause, as_clause=None):
        if as_clause is not None:
            import_clause += " as " + as_clause

        self._from_imports[from_clause].add(import_clause)

    def as_source_code(self):
        code = []

        for import_clause in sorted(self._imports):
            code.append(f"import {import_clause}\n")

        if self._imports:
            code.append("\n")

        for from_clause, import_clauses in sorted(self._from_imports.items()):
            if len(import_clauses) == 1:
                code.append(f"from {from_clause} import {list(import_clauses)[0]}\n")
            else:
                code.append(f"from {from_clause} import (\n")
                for clause in sorted(list(import_clauses)):
                    code.append(f"    {clause},\n")
                code.append(")\n")

        if self._from_imports:
            code.append("\n")

        return code


class CodeEditor:
    _filepath: str
    _lines: List[str]
    _locks: Dict[str, slice]
    _imports: ImportCollector

    def __init__(self, filepath):
        self._filepath = filepath
        self._lines = []
        self._locks = {}
        self._imports = ImportCollector()

    @property
    def locks(self):
        return self._locks.copy()

    @property
    def imports(self):
        return self._imports

    def _to_slice(self, key):
        if isinstance(key, str):
            key = self._locks[key]
        elif isinstance(key, int):
            key = slice(key, key + 1)
        elif not isinstance(key, slice):
            raise ValueError(f"Unknown key type {type(key)}")

        return key

    def __len__(self):
        return len(self._lines)

    def __str__(self):
        return f"CodeEditor({self._filepath})"

    def __repr__(self):
        return f'CodeEditor("{self._filepath}")'

    def __delitem__(self, key):
        self[key] = []

    def __getitem__(self, key):
        key = self._to_slice(key)

        return self._lines[key]

    def __setitem__(self, key, new_lines):
        if isinstance(key, str):
            name = key
            if name not in self._locks:
                key = slice(len(self), len(self))
        else:
            name = None

        key = self._to_slice(key)
        start = key.start if key.start is not None else 0
        stop = key.stop if key.stop is not None else len(self)

        contained = self._get_contained_locks(start, stop)
        for name, count in contained.items():
            if count == 1:
                raise ValueError(f"Lines {key} have intersection with lock {name}")

        for name in contained:
            del self._locks[name]

        if isinstance(new_lines, str):
            new_lines = [new_lines]

        self._lines = self._lines[:start] + new_lines + self._lines[stop:]
        self._shift_lock_boundaries(start, len(new_lines) - (stop - start))
        if name is not None:
            self.set_lock(name, start, start + len(new_lines), check=False)

    def __contains__(self, key):
        return key in self._locks

    def _shift_lock_boundaries(self, origin, diff):
        locks = {}
        for name, lock in self._locks.items():
            start = lock.start
            stop = lock.stop

            if start > origin:
                start += diff

            if stop > origin:
                stop += diff

            if start < stop:
                locks[name] = slice(start, stop)

        self._locks = locks

    def _get_contained_locks(self, start, stop):
        if start == stop:
            return {}

        counts = {}
        for name, lock in self._locks.items():
            if start == lock.start or stop == lock.stop:
                if start < lock.start or lock.stop < stop:
                    counts[name] = 2
            else:
                include_start = (start <= lock.start) and (lock.start < stop)
                include_end = (start < lock.stop) and (lock.stop <= stop)
                n_overlaps = include_start + include_end

                if n_overlaps > 0:
                    counts[name] = n_overlaps

        return counts

    def add_lines(self, *lines, lineno=None):
        if lineno is None:
            lineno = len(self)

        self[lineno:lineno] = list(lines)

    def set_lock(self, name, start, stop, check=True):
        if check:
            for lock in self._locks.values():
                # checking if new lock has intersection with an existing one
                if max(start, lock.start) <= min(stop, lock.stop):
                    raise RuntimeError(
                        "Failed to set the lock due to non-empty intersection with existing lock."
                    )
        self._locks[name] = slice(start, stop)

    @staticmethod
    def _get_indent(line: str):
        stripped = line.lstrip()
        if not stripped:
            return ""

        n_blanks = len(line) - len(stripped)

        return line[:n_blanks]

    def set_with_lock(
        self, name, lines, header_indent=None, footer_indent=None, lineno=None
    ):
        existed = name in self
        code = self.wrap_with_lock(
            name,
            lines,
            header_indent=header_indent,
            footer_indent=footer_indent,
        )
        if lineno is None:
            self[name] = code
        else:
            self[lineno:lineno] = code

        return not existed

    @staticmethod
    def wrap_with_lock(name, lines, header_indent=None, footer_indent=None):
        if header_indent is None:
            if lines:
                header_indent = CodeEditor._get_indent(lines[0])
            else:
                header_indent = ""

        if footer_indent is None:
            if lines:
                footer_indent = CodeEditor._get_indent(lines[-1])
            else:
                footer_indent = ""

        header = header_indent + LOCK_HEADER + name + LOCK_WARNING + "\n"
        footer = footer_indent + LOCK_FOOTER + "\n"

        return [header] + lines + [footer]

    def add_import(self, import_clause, as_clause=None):
        self._imports.add_import(import_clause, as_clause)

    def add_from_import(self, from_clause, import_clause, as_clause=None):
        self._imports.add_from_import(from_clause, import_clause, as_clause)

    def get_source_code(self):
        imports = self._imports.as_source_code()

        if "imports" in self:
            self.set_with_lock("imports", imports)
        elif self._imports:
            post_new_line = len(self._lines) > 0
            self.set_with_lock("imports", imports, lineno=0)

            if post_new_line:
                self.add_lines("\n", "\n", lineno=len(imports) + 2)

        return "".join(self._lines)

    def infer_locks(self):
        locks = {}
        name = None
        start = None
        for lineno, line in enumerate(self._lines):
            stripped = line.strip()
            if stripped.startswith(LOCK_HEADER):
                if start is not None:
                    raise RuntimeError("Nested locks are not supported.")

                if not stripped.endswith(LOCK_WARNING):
                    raise RuntimeError("Malformed lock header")

                start = lineno
                name = stripped[len(LOCK_HEADER) : -len(LOCK_WARNING)]
            elif stripped == LOCK_FOOTER:
                locks[name] = slice(start, lineno + 1)
                name = None
                start = None

        if start is not None:
            raise RuntimeError(f"Lock '{name}' never ended.")

        self._locks = locks

    def load(self, infer=True):
        if not os.path.exists(self._filepath):
            return

        with open(self._filepath, "r") as file:
            self._lines = file.readlines()

        if infer:
            self.infer_locks()

    def save(self):
        path = Path(self._filepath)
        path.parent.mkdir(parents=True, exist_ok=True)

        with open(path, "w") as file:
            file.write(self.get_source_code())



import argparse
from base64 import b64decode as b64d
from base58 import b58encode as b58e
from solana.rpc.api import Client
from solana.rpc.types import MemcmpOpts


def main():
    ap = argparse.ArgumentParser()
    ap.add_argument("dex_program_id")
    ap.add_argument("owner")
    ap.add_argument("--network", default="devnet")
    ap.add_argument("--market_product_group", default=None)
    args = ap.parse_args()

    urls = {
        "devnet": "https://api.devnet.solana.com",
        "dev": "https://api.devnet.solana.com",
        "localnet": "https://localhost:8899/",
        "local": "https://localhost:8899/",
        "mainnet": "https://api.mainnet-beta.solana.com/",
        "mainnet-beta": "https://api.mainnet-beta.solana.com/",
    }

    client = Client(urls[args.network])

    print(f"owner: {args.owner}")
    mem_cmp = [MemcmpOpts(8 + 33, args.owner)]
    if args.market_product_group != None:
        mem_cmp.append(MemcmpOpts(8 + 1, args.market_product_group))
        print(f"market product group: {args.market_product_group}")
    print("Searching...\n")
    resp = client.get_program_accounts(
        args.dex_program_id,
        "confirmed",
        encoding="base64",
        memcmp_opts=mem_cmp,
    )

    mpgs = {}
    for account in resp["result"]:
        data = b64d(account["account"]["data"][0])
        mpg = b58e(data[9:41]).decode('ascii')
        mpgs[mpg] = mpgs.get(mpg, []) + [account["pubkey"]]

    for k in mpgs:
        print(k)
        for trg in mpgs[k]:
            print(f"\t{trg}")


if __name__ == '__main__':
    main()


# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.instruments.types.account_tag import AccountTag
from dexteritysdk.codegen.instruments.types.expiration_status import ExpirationStatus
from podite import (
    FixedLenArray,
    U64,
    pod,
)
from solana.publickey import PublicKey
from dexteritysdk.solmate.dtypes import UnixTimestamp

# LOCK-END


# LOCK-BEGIN[class(FixedIncomeMetadata)]: DON'T MODIFY
@pod
class FixedIncomeMetadata:
    tag: AccountTag
    bump: U64
    face_value: U64
    coupon_rate: U64
    initialization_time: UnixTimestamp
    coupon_dates: FixedLenArray[UnixTimestamp, 32]
    maturity_date: UnixTimestamp
    market_product_group: PublicKey
    close_authority: PublicKey
    last_funding_time: UnixTimestamp
    expired: ExpirationStatus
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.open_orders_metadata import OpenOrdersMetadata
from dexteritysdk.codegen.dex.types.open_orders_node import OpenOrdersNode
from dexteritysdk.solmate.dtypes import Usize
from podite import (
    FixedLenArray,
    U64,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(OpenOrders)]: DON'T MODIFY
@pod
class OpenOrders:
    free_list_head: Usize
    total_open_orders: U64
    products: FixedLenArray[OpenOrdersMetadata, 256]
    orders: FixedLenArray["OpenOrdersNode", 1024]
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



from podite import U8, I32, pod, Enum
from solana.publickey import PublicKey
from solana.transaction import AccountMeta, TransactionInstruction


@pod
class InstructionCode(Enum[U8]):
    FindFees = 0
    InitializeTraderAcct = 1
    UpdateFees = 2


@pod
class NoParams:
    instr: InstructionCode


@pod
class UpdateFeesParams:
    instr: InstructionCode
    maker_fee_bps: I32
    taker_fee_bps: I32


def update_fees_ix(
    program_id: PublicKey,
    payer: PublicKey,
    fee_model_config_acct: PublicKey,
    market_product_group: PublicKey,
    system_program: PublicKey,
    maker_fee_bps: int,
    taker_fee_bps: int,
) -> TransactionInstruction:
    keys = [
        AccountMeta(pubkey=payer, is_signer=True, is_writable=False),
        AccountMeta(pubkey=fee_model_config_acct, is_signer=False, is_writable=True),
        AccountMeta(pubkey=market_product_group, is_signer=False, is_writable=False),
        AccountMeta(pubkey=system_program, is_signer=False, is_writable=False),
    ]
    return TransactionInstruction(
        keys=keys,
        program_id=program_id,
        data=UpdateFeesParams.to_bytes(UpdateFeesParams(
            instr=InstructionCode.UpdateFees,
            maker_fee_bps=maker_fee_bps,
            taker_fee_bps=taker_fee_bps,
        )),
    )


def initialize_trader_acct_ix(
    program_id: PublicKey,
    payer: PublicKey,
    fee_model_config_acct: PublicKey,
    trader_fee_acct: PublicKey,
    market_product_group: PublicKey,
    trader_risk_group: PublicKey,
    system_program: PublicKey,
) -> TransactionInstruction:
    keys = [
        AccountMeta(pubkey=payer, is_signer=True, is_writable=False),
        AccountMeta(pubkey=fee_model_config_acct, is_signer=False, is_writable=False),
        AccountMeta(pubkey=trader_fee_acct, is_signer=False, is_writable=True),
        AccountMeta(pubkey=market_product_group, is_signer=False, is_writable=False),
        AccountMeta(pubkey=trader_risk_group, is_signer=False, is_writable=False),
        AccountMeta(pubkey=system_program, is_signer=False, is_writable=False),
    ]
    return TransactionInstruction(
        program_id=program_id,
        keys=keys,
        data=NoParams.to_bytes(NoParams(instr=InstructionCode.InitializeTraderAcct)),
    )



# LOCK-BEGIN[imports]: DON'T MODIFY
from podite import (
    Enum,
    U64,
    pod,
)
from dexteritysdk.solmate.anchor import InstructionDiscriminant

# LOCK-END


# LOCK-BEGIN[instruction_tag]: DON'T MODIFY
@pod
class InstructionTag(Enum[U64]):
    VALIDATE_ACCOUNT_HEALTH = InstructionDiscriminant()
    VALIDATE_ACCOUNT_LIQUIDATION = InstructionDiscriminant()
    CREATE_RISK_STATE_ACCOUNT = InstructionDiscriminant()
    # LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.product import Product
from podite import (
    FixedLenArray,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(ProductArray)]: DON'T MODIFY
@pod
class ProductArray:
    array: FixedLenArray[Product, 256]
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.solmate.dtypes import Usize
from podite import (
    U64,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(OpenOrdersMetadata)]: DON'T MODIFY
@pod
class OpenOrdersMetadata:
    ask_qty_in_book: "Fractional"
    bid_qty_in_book: "Fractional"
    head_index: Usize
    num_open_orders: U64
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from .cancel_order import (
    CancelOrderIx,
    cancel_order,
)
from .choose_successor import (
    ChooseSuccessorIx,
    choose_successor,
)
from .claim_authority import (
    ClaimAuthorityIx,
    claim_authority,
)
from .clear_expired_orderbook import (
    ClearExpiredOrderbookIx,
    clear_expired_orderbook,
)
from .consume_orderbook_events import (
    ConsumeOrderbookEventsIx,
    consume_orderbook_events,
)
from .deposit_funds import (
    DepositFundsIx,
    deposit_funds,
)
from .initialize_combo import (
    InitializeComboIx,
    initialize_combo,
)
from .initialize_market_product import (
    InitializeMarketProductIx,
    initialize_market_product,
)
from .initialize_market_product_group import (
    InitializeMarketProductGroupIx,
    initialize_market_product_group,
)
from .initialize_print_trade import (
    InitializePrintTradeIx,
    initialize_print_trade,
)
from .initialize_trader_risk_group import (
    InitializeTraderRiskGroupIx,
    initialize_trader_risk_group,
)
from .instruction_tag import InstructionTag
from .new_order import (
    NewOrderIx,
    new_order,
)
from .remove_market_product import (
    RemoveMarketProductIx,
    remove_market_product,
)
from .remove_market_product_group import (
    RemoveMarketProductGroupIx,
    remove_market_product_group,
)
from .setup_capital_limits import (
    SetupCapitalLimitsIx,
    setup_capital_limits,
)
from .sign_print_trade import (
    SignPrintTradeIx,
    sign_print_trade,
)
from .sweep_fees import (
    SweepFeesIx,
    sweep_fees,
)
from .transfer_full_position import (
    TransferFullPositionIx,
    transfer_full_position,
)
from .update_capital_limits import (
    UpdateCapitalLimitsIx,
    update_capital_limits,
)
from .update_market_product_group import (
    UpdateMarketProductGroupIx,
    update_market_product_group,
)
from .update_product_funding import (
    UpdateProductFundingIx,
    update_product_funding,
)
from .update_trader_funding import (
    UpdateTraderFundingIx,
    update_trader_funding,
)
from .update_variance_cache import (
    UpdateVarianceCacheIx,
    update_variance_cache,
)
from .withdraw_funds import (
    WithdrawFundsIx,
    withdraw_funds,
)

# LOCK-END



from typing import Type

from podite import U32, U64, I64, Enum, Variant, Option, pod
from podite._utils import _GetitemToCall, get_calling_module


Usize = U64  # Should it be U32?
UnixTimestamp = I64


class Error(Variant):
    msg: str

    def __init__(self, /, value=None, field=None, module=None, msg=None):
        super().__init__(value, field, module)

        self.msg = msg

    def to_string(self, instance):
        msg = self.msg.format(instance.field)
        return f"{self.name}: {msg}"


class ProgramError(Enum):
    CUSTOM = Error(field=U32, msg="Custom program error: {}")
    INVALID_ARGUMENT = Error(
        msg="The arguments provided to a program instruction where invalid"
    )
    INVALID_INSTRUCTION_DATA = Error(msg="An instruction's data contents was invalid")
    INVALID_ACCOUNT_DATA = Error(msg="An account's data contents was invalid")
    ACCOUNT_DATA_TOO_SMALL = Error(msg="An account's data was too small")
    INSUFFICIENT_FUNDS = Error(
        msg="An account's balance was too small to complete the instruction"
    )
    INCORRECT_PROGRAM_ID = Error(msg="The account did not have the expected program id")
    MISSING_REQUIRED_SIGNATURE = Error(msg="A signature was required but not found")
    ACCOUNT_ALREADY_INITIALIZED = Error(
        msg="An initialize instruction was sent to an account that has already been initialized"
    )
    UNINITIALIZED_ACCOUNT = Error(
        msg="An attempt to operate on an account that hasn't been initialized"
    )
    NOT_ENOUGH_ACCOUNT_KEYS = Error(
        msg="The instruction expected additional account keys"
    )
    ACCOUNT_BORROW_FAILED = Error(msg="Failed to borrow a reference to account data")
    MAX_SEED_LENGTH_EXCEED = Error(
        msg="Length of the seed is too long for address generation"
    )
    INVALID_SEEDS = Error(msg="Provided seeds do not result in a valid address")
    BORCH_IO_ERROR = Error(field=str, msg="IO Error: {}")
    ACCOUNT_NOT_RENT_EXEMPT = Error(
        msg="An account does not have enough lamports to be rent-exempt"
    )
    UNSUPPORTED_SYSVAR = Error(msg="Unsupported sysvar")
    ILLEGAL_OWNER = Error(msg="Provided owner is not allowed")
    ACCOUNTS_DATA_BUDGET_EXCEEDED = Error(
        msg="Requested account data allocation exceeded the accounts data budget"
    )

    def __str__(self):
        variant = self._get_variant(self.get_name())
        return variant.to_string(self)


def _coption(name, type_: Type):
    @pod
    class _COption(Enum[U32]):
        NONE = Variant()
        SOME = Variant(field=type_, module=get_calling_module(4))

    _COption.__name__ = f"{name}[{type_}]"
    _COption.__qualname__ = _COption.__name__

    return _COption


COption = _GetitemToCall("COption", _coption)



from enum import IntEnum

from podite import U8, Enum, pod


@pod
class InstructionCode(Enum[U8]):
    INITIALIZE_CLOCK = None
    INITIALIZE_ORACLE = None
    UPDATE_CLOCK = None
    UPDATE_PRICE = None



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.solmate.dtypes import Usize
from dexteritysdk.utils.aob.state.base import Side
from podite import pod

# LOCK-END


# LOCK-BEGIN[class(InitializePrintTradeParams)]: DON'T MODIFY
@pod
class InitializePrintTradeParams:
    product_index: Usize
    size: Fractional
    price: Fractional
    side: Side
    operator_creator_fee_proportion: Fractional
    operator_counterparty_fee_proportion: Fractional
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



import os
from dexteritysdk.local_env import *
import sys
import json
import time
import argparse
import base64

from pathlib import Path

import solana.system_program as sp

from solana.rpc.api import Client
from solana.keypair import Keypair
from solana.publickey import PublicKey
from solana.rpc.commitment import Confirmed
from solana.rpc import types
from solana.transaction import Transaction
from dexteritysdk.utils.solana import get_global_parser

import dexteritysdk.dex.actions as dex_act
import dexteritysdk.constant_fees.actions as fee_act
import dexteritysdk.instruments.actions as ins_act
from dexteritysdk.utils.solana import explore, AccountParser
from dexteritysdk.dex.state import account_parser as dex_account_parser
from dexteritysdk.dex.state import ProductStatus, MarketProductGroup, TraderRiskGroup
from dexteritysdk.utils.aob import account_parser as aob_account_parser
from dexteritysdk.instruments.state import account_parser as ins_account_parser
from dexteritysdk.utils.random_hash import random_hash
from dexteritysdk.utils.solana import (
    set_global_client,
    set_signers,
    send_transaction,
    get_global_client,
    get_signers,
)
from dexteritysdk.program_ids import DEX_PROGRAM_ID, AOB_PROGRAM_ID, INSTRUMENTS_PROGRAM_ID


RPC_URL = "https://api.devnet.solana.com"

SUCCESS = "\u2705"


# TODO: move this somewhere better
def query_all_trader_risk_groups(market_product_group: PublicKey):
    client = get_global_client()
    raw_dex_accounts = client.get_program_accounts(
        pubkey=DEX_PROGRAM_ID,
        encoding="base64",
        data_size=TraderRiskGroup.calc_size() + 8,  # currently need constant offset
    )[
        "result"
    ]  # note: this shouldn't have a cap on the number returned right?

    trader_risk_keys = []
    trader_risk_groups = []
    for acct in raw_dex_accounts:
        key = PublicKey(acct["pubkey"])
        trg: TraderRiskGroup = dex_account_parser(
            base64.b64decode(acct["account"]["data"][0])
        )
        if str(trg.market_product_group) == str(market_product_group):
            trader_risk_keys.append(key)
            trader_risk_groups.append(trg)

    return trader_risk_keys, trader_risk_groups


def expiration_routine(auth: PublicKey, mpg_key: PublicKey, product_key: PublicKey):
    mpg_obj: MarketProductGroup = explore(mpg_key).data_obj
    assert isinstance(mpg_obj, MarketProductGroup)
    product_idx = mpg_obj.find_market_index(product_key)
    product_obj = mpg_obj.get_product_by_key(product_key)
    assert product_obj is not None
    combo_objs = mpg_obj.get_combos_for_outright(product_key)
    combo_keys = [c.product_key for c in combo_objs]

    # optional: actually call settle on the instrument
    assert product_obj.product_status == ProductStatus.EXPIRED

    # crank all event queue events (outright and combo) -> all relevant event queues should be empty
    # if we assume that there is an event queue crank running for the market, can just confirm here!
    for key_i in combo_keys + [product_key]:
        product_i = mpg_obj.mpg_obj.get_product_by_key(
            key_i
        ) or mpg_obj.get_combo_by_key(key_i)
        orderbook = explore(product_i.orderbook).data_obj
        event_queue_key = orderbook.event_queue
        event_queue = explore(event_queue_key).data_obj
        while event_queue.header.count > 0:
            _, resp = dex_act.consume_orderbook_events(
                market_product_group=mpg_key,
                product_key=key_i,
                reward_target=auth,
                max_iterations=10,
            )
            # reload event queue to check when empty
            time.sleep(3)
            event_queue = explore(event_queue_key).data_obj

    # query all trader_risk_groups, call settle on each that has open position -> open interest should be zero
    # should be able to confirm that long and short open interest are both zero at the end of this step
    while product_obj.open_long_interest > 0 or product_obj.open_short_interest > 0:
        all_trader_keys, all_trader_risk_groups = query_all_trader_risk_groups(mpg_key)
        for trg_key, trg in zip(all_trader_keys, all_trader_risk_groups):
            if trg.is_active_product(product_idx):
                _, resp = dex_act.update_trader_funding(
                    market_product_group=mpg_key,
                    trader=trg_key,
                    fee_payer=auth,
                )
        # reload objects to confirm that everyone settled
        time.sleep(5)
        mpg_obj = explore(mpg_key).data_obj
        product_obj = mpg_obj.get_product_by_key(product_key)
        print(
            f"product open interest: ({product_obj.open_long_interest}, {product_obj.open_short_interest})"
        )

    # for each combo and product, crank until all open orders canceled and then remove the product from market product group
    for key_i in combo_keys + [product_key]:
        while True:
            _, resp = dex_act.clear_expired_orders(
                market_product_group=mpg_key,
                product_key=key_i,
                authority=auth,
                num_orders_to_cancel=10,
            )

        _, resp = dex_actions.remove_market_product(
            market_product_group=mpg,
            product_key=key_i,
        )


def main():
    curr_directory = os.path.dirname(os.path.abspath(__file__))
    cfg_path = os.path.join(curr_directory, "config.json")
    if os.path.exists(cfg_path):
        with open(cfg_path) as f:
            cfg = json.load(f)
            mpg_pubkey = cfg["mpg_key"]
    else:
        ap = argparse.ArgumentParser()
        ap.add_argument("mpg_pubkey")
        args = ap.parse_args()
        mpg_pubkey = args.mpg_pubkey

    auth = Keypair(b"nima" * 8)
    market_product_group_pubkey = PublicKey(mpg_pubkey)
    product_key = PublicKey()

    parser = AccountParser()
    parser.register_parser(DEX_PROGRAM_ID, dex_account_parser)
    parser.register_parser(AOB_PROGRAM_ID, aob_account_parser)
    parser.register_parser(INSTRUMENTS_PROGRAM_ID.ins_account_parser)

    client = Client(RPC_URL)
    set_global_client(client)

    set_signers(auth)

    expiration_routine(auth.public_key, market_product_group_pubkey, product_key)


if __name__ == "__main__":
    main()



from solana.publickey import PublicKey
from podite import U64, pod, FixedLenArray


@pod
class MarketState:
    tag: U64
    caller_authority: PublicKey
    event_queue: PublicKey
    bids: PublicKey
    asks: PublicKey
    callback_id_len: U64
    callback_info_len: U64
    fee_budget: U64
    initial_lamports: U64
    min_base_order_size: U64
    price_bitmask: U64
    cranker_reward: U64



import os
import json
import re
import subprocess
from functools import partial
from typing import Callable, Dict, Union, Set

from solana.publickey import PublicKey
from solana.system_program import SYS_PROGRAM_ID
from solana.sysvar import SYSVAR_RENT_PUBKEY
from dexteritysdk.solmate.anchor import Idl
from dexteritysdk.solmate.anchor.codegen import usize_type, unix_timestamp_type, program_error_type, CodeGen
from dexteritysdk.solmate.anchor.editor import CodeEditor
from dexteritysdk.solmate.utils import pascal_to_snake
from spl.token.constants import TOKEN_PROGRAM_ID

from dexteritysdk.scripts.extract_program_ids import get_root


def main():
    root = get_root()
    cli(idl_dir=f'{root}/target/idl', out_dir=f'{root}/client', pids_dir=f'{root}/target/deploy',
        parent_module='dexteritysdk.codegen', skip_types={"TwoIterators", "DomainOrProgramError"})


if __name__ == '__main__':
    main()


def side_type(editor: CodeEditor):
    editor.add_from_import("dexteritysdk.utils.aob.state.base", "Side")
    return "Side"


def self_trade_behavior_type(editor: CodeEditor):
    editor.add_from_import("dexteritysdk.utils.aob.state.base", "SelfTradeBehavior")
    return "SelfTradeBehavior"


def cli(
        idl_dir: str,
        out_dir: str,
        pids_dir: Union[str, Dict[str, PublicKey]],
        parent_module: str,
        skip_types: Set[str],
):
    protocol_to_idl_and_types = {}
    protocol_to_pid = get_protocols(idl_dir, pids_dir)
    for protocol in protocol_to_pid.keys():
        idl = Idl.from_json_file(f"{idl_dir}/{protocol}.json")

        idl.types = list(filter(lambda x: x.name not in skip_types, idl.types))
        idl.accounts = list(filter(lambda x: x.name not in skip_types, idl.accounts))
        protocol_to_idl_and_types[protocol] = (
            idl,
            defined_types_to_imports(f"{parent_module}.{protocol}", idl),
        )

        if protocol.upper() in os.environ:
            env_pid = os.environ[protocol.upper()]
            print(f"Using program_id from ENV for {protocol}: {env_pid}")
            protocol_to_pid[protocol] = env_pid
        else:
            print(f"Using program_id from target/deploy/{protocol}-keypair.json for {protocol}: "
                  f"{protocol_to_pid[protocol]}")

        if protocol_to_pid[protocol] is None:
            raise Exception("Found IDL but no program id set")

    external_types = {
        "usize": usize_type,
        "UnixTimestamp": unix_timestamp_type,
        "ProgramError": program_error_type,
        "SelfTradeBehavior": self_trade_behavior_type,
        "Side": side_type,
    }
    # allow each IDL to reference types defined in other IDLs
    for (_, exported_types) in protocol_to_idl_and_types.values():
        external_types.update(exported_types)

    for protocol, (idl, _) in protocol_to_idl_and_types.items():
        print(f"Generating code for {protocol}")
        codegen = CodeGen(
            idl=idl,
            addresses={},
            root_module=f"{parent_module}.{protocol}",
            source_path=out_dir,
            external_types=external_types,
            default_accounts={
                "systemProgram": SYS_PROGRAM_ID,
                "token_program": TOKEN_PROGRAM_ID,
                "sysvar_rent": SYSVAR_RENT_PUBKEY,
                "program_id": protocol_to_pid[protocol],
            },
            accnt_tag_values="anchor",
            instr_tag_values="anchor",
            skip_types=skip_types,
        )
        codegen.generate_code(check_missing_types=not True)
        codegen.save_modules()


def defined_types_to_imports(
        root_module: str, idl: Idl
) -> Dict[str, Callable[[CodeEditor], str]]:
    def add_import(name: str, editor: CodeEditor) -> str:
        editor.add_from_import(f"{root_module}.types.{pascal_to_snake(name)}", name)
        return name

    type_definitions = idl.types + idl.accounts
    return dict(((ty.name, partial(add_import, ty.name)) for ty in type_definitions))


def get_protocols(idl_dir: str, pids: Union[str, Dict[str, str]]) -> Dict[str, str]:
    protocols = set()
    for filename in os.listdir(idl_dir):
        match = re.search(r"([a-z_\-]+).json", filename)
        if match is None:
            continue
        protocol = match.groups()[0]
        protocols.add(protocol)

    # if a path was passed in, load pids from the directory
    if isinstance(pids, str):
        pids = dir_to_pids(pids)

    intersection = {}
    for protocol, pid in pids.items():
        if protocol in protocols:
            intersection[protocol] = pid
    for protocol in protocols:
        if protocol not in pids:
            print("WARNING: found idl file with no matching program id: ", protocol)
            intersection[protocol] = None

    return intersection


def dir_to_pids(dir: str) -> Dict[str, str]:
    program_to_id = {}
    root = run("git rev-parse --show-toplevel")
    for filename in os.listdir(dir):
        match = re.search(r"([a-z_]+)-keypair.json", filename)
        if match is None:
            continue
        program = match.groups()[0]
        program_to_id[program] = run(f"solana-keygen pubkey {dir}/{filename}")
    return program_to_id


def run(cmd, debug=False):
    if debug:
        print(cmd)
    res = subprocess.check_output(cmd, shell=True).strip().decode("utf-8")
    if debug:
        print(res)
    return res



import struct

from solana.publickey import PublicKey

from dexteritysdk.codegen.dex.types import Fractional
from dexteritysdk.program_ids import INSTRUMENTS_PROGRAM_ID
from dexteritysdk.codegen.instruments.types import InstrumentType


def get_derivative_key(
    price_oracle: PublicKey,
    market_product_group: PublicKey,
    instrument_type: InstrumentType,
    strike: float,
    full_funding_period: int,
    minimum_funding_period: int,
    initialization_time: int,
    **kwargs,
):
    strike = Fractional.to_decimal(strike)  # type: Fractional
    derivative_metadata, bump_seed = PublicKey.find_program_address(
        seeds=[
            b"derivative",
            bytes(price_oracle),
            bytes(market_product_group),
            struct.pack("<Q", int(instrument_type)),  # fix this to_bytes()
            struct.pack("<q", strike.m),
            struct.pack("<Q", strike.exp),
            struct.pack("<q", initialization_time),
            struct.pack("<q", full_funding_period),
            struct.pack("<q", minimum_funding_period),
        ],
        program_id=INSTRUMENTS_PROGRAM_ID,
    )
    return derivative_metadata, bump_seed



# LOCK-BEGIN[imports]: DON'T MODIFY
from .instruction_tag import InstructionTag
from dataclasses import dataclass
from dexteritysdk.codegen.instruments.types import InitializeDerivativeParams
from dexteritysdk.solmate.utils import to_account_meta
from io import BytesIO
from podite import BYTES_CATALOG
from solana.publickey import PublicKey
from solana.transaction import (
    AccountMeta,
    TransactionInstruction,
)
from typing import (
    List,
    Optional,
    Union,
)

# LOCK-END


# LOCK-BEGIN[ix_cls(initialize_derivative)]: DON'T MODIFY
@dataclass
class InitializeDerivativeIx:
    program_id: PublicKey

    # account metas
    derivative_metadata: AccountMeta
    price_oracle: AccountMeta
    market_product_group: AccountMeta
    payer: AccountMeta
    system_program: AccountMeta
    clock: AccountMeta
    remaining_accounts: Optional[List[AccountMeta]]

    # data fields
    params: InitializeDerivativeParams

    def to_instruction(self):
        keys = []
        keys.append(self.derivative_metadata)
        keys.append(self.price_oracle)
        keys.append(self.market_product_group)
        keys.append(self.payer)
        keys.append(self.system_program)
        keys.append(self.clock)
        if self.remaining_accounts is not None:
            keys.extend(self.remaining_accounts)

        buffer = BytesIO()
        buffer.write(InstructionTag.to_bytes(InstructionTag.INITIALIZE_DERIVATIVE))
        buffer.write(BYTES_CATALOG.pack(InitializeDerivativeParams, self.params))

        return TransactionInstruction(
            keys=keys,
            program_id=self.program_id,
            data=buffer.getvalue(),
        )

# LOCK-END


# LOCK-BEGIN[ix_fn(initialize_derivative)]: DON'T MODIFY
def initialize_derivative(
    derivative_metadata: Union[str, PublicKey, AccountMeta],
    price_oracle: Union[str, PublicKey, AccountMeta],
    market_product_group: Union[str, PublicKey, AccountMeta],
    payer: Union[str, PublicKey, AccountMeta],
    clock: Union[str, PublicKey, AccountMeta],
    params: InitializeDerivativeParams,
    system_program: Union[str, PublicKey, AccountMeta] = PublicKey("11111111111111111111111111111111"),
    remaining_accounts: Optional[List[AccountMeta]] = None,
    program_id: Optional[PublicKey] = None,
):
    if program_id is None:
        program_id = PublicKey("8981bZYszfz1FrFVx7gcUm61RfawMoAHnURuERRJKdkq")

    if isinstance(derivative_metadata, (str, PublicKey)):
        derivative_metadata = to_account_meta(
            derivative_metadata,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(price_oracle, (str, PublicKey)):
        price_oracle = to_account_meta(
            price_oracle,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(market_product_group, (str, PublicKey)):
        market_product_group = to_account_meta(
            market_product_group,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(payer, (str, PublicKey)):
        payer = to_account_meta(
            payer,
            is_signer=True,
            is_writable=True,
        )
    if isinstance(system_program, (str, PublicKey)):
        system_program = to_account_meta(
            system_program,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(clock, (str, PublicKey)):
        clock = to_account_meta(
            clock,
            is_signer=False,
            is_writable=False,
        )

    return InitializeDerivativeIx(
        program_id=program_id,
        derivative_metadata=derivative_metadata,
        price_oracle=price_oracle,
        market_product_group=market_product_group,
        payer=payer,
        system_program=system_program,
        clock=clock,
        remaining_accounts=remaining_accounts,
        params=params,
    ).to_instruction()

# LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
from podite import (
    AutoTagType,
    Enum,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(AccountTag)]: DON'T MODIFY
@pod
class AccountTag(Enum[AutoTagType]):
    UNINITIALIZED = None
    MARKET_PRODUCT_GROUP = None
    TRADER_RISK_GROUP = None
    TRADER_POSITION = None
    MARKET_PRODUCT_GROUP_WITH_COMBOS = None
    COMBO_GROUP = None
    COMBO = None
    RISK_PROFILE = None
    # LOCK-END

    @classmethod
    def _to_bytes_partial(cls, buffer, obj, **kwargs):
        # to modify packing, change this method
        return super()._to_bytes_partial(buffer, obj, **kwargs)

    @classmethod
    def _from_bytes_partial(cls, buffer, **kwargs):
        # to modify unpacking, change this method
        return super()._from_bytes_partial(buffer, **kwargs)

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from .instruction_tag import InstructionTag
from dataclasses import dataclass
from dexteritysdk.solmate.utils import to_account_meta
from io import BytesIO
from podite import BYTES_CATALOG
from solana.publickey import PublicKey
from solana.transaction import (
    AccountMeta,
    TransactionInstruction,
)
from typing import (
    List,
    Optional,
    Union,
)

# LOCK-END


# LOCK-BEGIN[ix_cls(transfer_full_position)]: DON'T MODIFY
@dataclass
class TransferFullPositionIx:
    program_id: PublicKey

    # account metas
    liquidator: AccountMeta
    market_product_group: AccountMeta
    liquidatee_risk_group: AccountMeta
    liquidator_risk_group: AccountMeta
    risk_engine_program: AccountMeta
    risk_model_configuration_acct: AccountMeta
    risk_output_register: AccountMeta
    liquidator_risk_state_account_info: AccountMeta
    liquidatee_risk_state_account_info: AccountMeta
    risk_signer: AccountMeta
    remaining_accounts: Optional[List[AccountMeta]]

    def to_instruction(self):
        keys = []
        keys.append(self.liquidator)
        keys.append(self.market_product_group)
        keys.append(self.liquidatee_risk_group)
        keys.append(self.liquidator_risk_group)
        keys.append(self.risk_engine_program)
        keys.append(self.risk_model_configuration_acct)
        keys.append(self.risk_output_register)
        keys.append(self.liquidator_risk_state_account_info)
        keys.append(self.liquidatee_risk_state_account_info)
        keys.append(self.risk_signer)
        if self.remaining_accounts is not None:
            keys.extend(self.remaining_accounts)

        buffer = BytesIO()
        buffer.write(InstructionTag.to_bytes(InstructionTag.TRANSFER_FULL_POSITION))

        return TransactionInstruction(
            keys=keys,
            program_id=self.program_id,
            data=buffer.getvalue(),
        )

# LOCK-END


# LOCK-BEGIN[ix_fn(transfer_full_position)]: DON'T MODIFY
def transfer_full_position(
    liquidator: Union[str, PublicKey, AccountMeta],
    market_product_group: Union[str, PublicKey, AccountMeta],
    liquidatee_risk_group: Union[str, PublicKey, AccountMeta],
    liquidator_risk_group: Union[str, PublicKey, AccountMeta],
    risk_engine_program: Union[str, PublicKey, AccountMeta],
    risk_model_configuration_acct: Union[str, PublicKey, AccountMeta],
    risk_output_register: Union[str, PublicKey, AccountMeta],
    liquidator_risk_state_account_info: Union[str, PublicKey, AccountMeta],
    liquidatee_risk_state_account_info: Union[str, PublicKey, AccountMeta],
    risk_signer: Union[str, PublicKey, AccountMeta],
    remaining_accounts: Optional[List[AccountMeta]] = None,
    program_id: Optional[PublicKey] = None,
):
    if program_id is None:
        program_id = PublicKey("FUfpR31LmcP1VSbz5zDaM7nxnH55iBHkpwusgrnhaFjL")

    if isinstance(liquidator, (str, PublicKey)):
        liquidator = to_account_meta(
            liquidator,
            is_signer=True,
            is_writable=False,
        )
    if isinstance(market_product_group, (str, PublicKey)):
        market_product_group = to_account_meta(
            market_product_group,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(liquidatee_risk_group, (str, PublicKey)):
        liquidatee_risk_group = to_account_meta(
            liquidatee_risk_group,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(liquidator_risk_group, (str, PublicKey)):
        liquidator_risk_group = to_account_meta(
            liquidator_risk_group,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(risk_engine_program, (str, PublicKey)):
        risk_engine_program = to_account_meta(
            risk_engine_program,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(risk_model_configuration_acct, (str, PublicKey)):
        risk_model_configuration_acct = to_account_meta(
            risk_model_configuration_acct,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(risk_output_register, (str, PublicKey)):
        risk_output_register = to_account_meta(
            risk_output_register,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(liquidator_risk_state_account_info, (str, PublicKey)):
        liquidator_risk_state_account_info = to_account_meta(
            liquidator_risk_state_account_info,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(liquidatee_risk_state_account_info, (str, PublicKey)):
        liquidatee_risk_state_account_info = to_account_meta(
            liquidatee_risk_state_account_info,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(risk_signer, (str, PublicKey)):
        risk_signer = to_account_meta(
            risk_signer,
            is_signer=False,
            is_writable=False,
        )

    return TransferFullPositionIx(
        program_id=program_id,
        liquidator=liquidator,
        market_product_group=market_product_group,
        liquidatee_risk_group=liquidatee_risk_group,
        liquidator_risk_group=liquidator_risk_group,
        risk_engine_program=risk_engine_program,
        risk_model_configuration_acct=risk_model_configuration_acct,
        risk_output_register=risk_output_register,
        liquidator_risk_state_account_info=liquidator_risk_state_account_info,
        liquidatee_risk_state_account_info=liquidatee_risk_state_account_info,
        risk_signer=risk_signer,
        remaining_accounts=remaining_accounts,
    ).to_instruction()

# LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from podite import (
    FixedLenArray,
    U64,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(PriceEwma)]: DON'T MODIFY
@pod
class PriceEwma:
    ewma_bid: FixedLenArray["Fractional", 4]
    ewma_ask: FixedLenArray["Fractional", 4]
    bid: "Fractional"
    ask: "Fractional"
    slot: U64
    prev_bid: "Fractional"
    prev_ask: "Fractional"
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



import json
import os
import re
import subprocess


def main():
    programs = get_program_to_id()
    print("Programs:")
    for name, id in programs.items():
        print(f"{name}: {id}")


def get_program_to_id():
    root = get_root()
    pids = dir_to_pids(f"{root}/target/deploy")
    return pids


def dir_to_pids(dir: str):
    program_to_id = {}
    for filename in os.listdir(dir):
        match = re.search(r"([a-z_]+)-keypair.json", filename)
        if match is None:
            continue
        program = match.groups()[0]
        program_to_id[program] = run(f"solana-keygen pubkey {dir}/{filename}")
    return program_to_id


def run(cmd, debug=False):
    if debug:
        print(cmd)
    res = subprocess.check_output(cmd, shell=True).strip().decode("utf-8")
    if debug:
        print(res)
    return res


def get_root() -> str:
    return run("git rev-parse --show-toplevel")


if __name__ == "__main__":
    main()



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.combo import Combo
from dexteritysdk.codegen.dex.types.outright import Outright
from podite import (
    AutoTagType,
    Enum,
    Option,
    Variant,
    named_fields,
    pod,
)

# LOCK-END

import dexteritysdk.codegen.dex.types as types


# LOCK-BEGIN[class(Product)]: DON'T MODIFY
@pod
class Product(Enum[AutoTagType]):
    OUTRIGHT = Variant(field=named_fields(outright=Outright))
    COMBO = Variant(field=named_fields(combo=Combo))
    # LOCK-END

    def is_combo(self) -> bool:
        return getattr(self.field, "combo", None) is not None

    def is_outright(self) -> bool:
        return getattr(self.field, "outright", None) is not None

    def metadata(self) -> "types.ProductMetadata":
        if getattr(self.field, "outright", None):
            return self.field.outright.metadata
        elif getattr(self.field, "combo", None):
            return self.field.combo.metadata
        else:
            raise ValueError(f"Uknown product type for {self}")

    @classmethod
    def _to_bytes_partial(cls, buffer, obj, **kwargs):
        # to modify packing, change this method
        return super()._to_bytes_partial(buffer, obj, **kwargs)

    @classmethod
    def _from_bytes_partial(cls, buffer, **kwargs):
        # to modify unpacking, change this method
        return super()._from_bytes_partial(buffer, **kwargs)

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



import argparse
import json

from base58 import b58encode as b58e
from solana.keypair import Keypair

def main():
    ap = argparse.ArgumentParser()
    ap.add_argument("f")
    args = ap.parse_args()

    with open(args.f, "r") as f:
        kp = json.load(f)
    print(b58e(bytes(kp)).decode('ascii'))
    print(Keypair(kp[:32]).public_key)

if __name__ == '__main__':
    main()



from .common import (
    InstructionCode,
)

from .initialize_clock import (
    initialize_clock_ix,
)

from .initialize_oracle import (
    initialize_oracle_ix,
)

from .update_clock import (
    update_clock_ix,
)

from .update_price import (
    update_price_ix,
)

__all__ = [
    "InstructionCode",
    "initialize_clock_ix",
    "initialize_oracle_ix",
    "update_clock_ix",
    "update_price_ix",
]



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from podite import (
    U8,
    pod,
)
from solana.publickey import PublicKey

# LOCK-END


# LOCK-BEGIN[class(CapitalLimits)]: DON'T MODIFY
@pod
class CapitalLimits:
    deposit_limit: Fractional
    withdrawal_limit: Fractional
    market_product_group: PublicKey
    bump: U8
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



from dexteritysdk.scripts.extract_program_ids import get_program_to_id, get_root
import os
import subprocess


def main():
    programs = get_program_to_id()
    root = get_root()
    cmd = "solana-test-validator"
    args = ""
    for (name, pid) in programs.items():
        path = f"{root}/target/deploy/{name}.so"
        args += f" --bpf-program {pid} {path}"
    print(f"Running {cmd} {args}")
    print(os.system(cmd + " " + args))


if __name__ == "__main__":
    main()



# LOCK-BEGIN[imports]: DON'T MODIFY
import dexteritysdk.codegen.instruments.accounts as accounts
import dexteritysdk.codegen.instruments.instructions as instructions
import dexteritysdk.codegen.instruments.types as types

# LOCK-END




# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.codegen.instruments.types.instrument_type import InstrumentType
from dexteritysdk.codegen.instruments.types.oracle_type import OracleType
from dexteritysdk.solmate.dtypes import UnixTimestamp
from podite import pod
from solana.publickey import PublicKey

# LOCK-END


# LOCK-BEGIN[class(InitializeDerivativeParams)]: DON'T MODIFY
@pod
class InitializeDerivativeParams:
    instrument_type: "InstrumentType"
    strike: Fractional
    full_funding_period: UnixTimestamp
    minimum_funding_period: UnixTimestamp
    initialization_time: UnixTimestamp
    close_authority: PublicKey
    oracle_type: "OracleType"
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



import argparse
import os

from dexteritysdk.solmate.anchor import codegen

TAG_TYPES = [
    "anchor",
    "incremental",
    "incremental:U8",
    "incremental:U16",
    "incremental:U32",
    "incremental:U64",
    "incremental:U128",
]


def default_account_gen(name_, value_):
    def _default_account(editor):
        from_clause, import_clause = value_.rsplit(".", maxsplit=1)
        as_clause = name_.upper()
        if as_clause == import_clause:
            editor.add_from_import(from_clause, import_clause)
        else:
            editor.add_from_import(from_clause, import_clause, name_.upper())
        return name_.upper()

    return _default_account


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--idl", type=str, required=True,
                        help="Path to idl file")
    parser.add_argument("--addrs", nargs="+", type=str, required=True, default=[])  # TODO: @nimily how does this work? Can you correct the types and add a help message
    parser.add_argument("--root-dir", type=str, required=True, default=os.getcwd(),
                        help="Path to output")
    parser.add_argument("--module", type=str, required=True,
                        help="Name of the python module")
    parser.add_argument("--skip-types", nargs="+", required=False, default=[],
                        help="Idl types to skip emitting code for")
    parser.add_argument(
        "--default-accounts",
        nargs="+",
        type=str,
        required=False,
        default=[],
        help="Dictionary of account name to public key which are used to enable default arguments. "
             "Must include program_id")
    parser.add_argument(
        "--instruction-tag",
        choices=TAG_TYPES,
        default="anchor",
        required=False,
        help="Type of instruction tag to use. Use 'anchor' if generating from an anchor idl",
    )
    parser.add_argument(
        "--account-tag",
        choices=TAG_TYPES,
        default="anchor",
        required=False,
        help="Type of account tag to use. Use 'anchor' if generating from an anchor idl",
    )
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()

    skip_types = set(args.skip_types)

    default_accounts = {}
    for acct in args.default_accounts:
        name, value = acct.split("=")
        if "." in value:
            default_accounts[name] = default_account_gen(name, value)
        else:
            default_accounts[name] = value

    addresses = {}
    for acct in args.addrs:
        name, value = acct.split("=")
        addresses[name] = value

    instruction_tag = args.instruction_tag
    if instruction_tag == "incremental":
        instruction_tag = "incremental:U8"

    account_tag = args.account_tag
    if account_tag == "incremental":
        account_tag = "incremental:U8"

    codegen.cli(
        args.idl,
        addresses,
        args.root_dir,
        args.module,
        skip_types,
        default_accounts,
        instruction_tag,
        account_tag,
    )


if __name__ == "__main__":
    main()



import struct
from numpy import isin
import pandas as pd
from enum import IntEnum
from dataclasses import field
from typing import Union
from io import BytesIO

from solana.publickey import PublicKey


from podite import (
    U8,
    pod,
    U32,
    U64,
    U128,
    Enum,
    Option,
    FixedLenArray,
    BYTES_CATALOG,
)

from .base import AccountTag, Side

NODE_SIZE = 32
FREE_NODE_SIZE = 4

NODE_TAG_SIZE = 8
SLOT_SIZE = NODE_TAG_SIZE + NODE_SIZE

SLAB_HEADER_LEN = 97
PADDED_SLAB_HEADER_LEN = SLAB_HEADER_LEN + 7

BINARY_ORDER_SCALE = 32


@pod
class InnerNode:
    prefix_len: U64
    key: U128
    children: FixedLenArray[U32, 2]


@pod
class LeafNode:
    key: U128
    callback_info_pt: U64
    base_quantity: U64

    @property
    def price(self):
        return self.key >> (64 + BINARY_ORDER_SCALE)

    @property
    def order_id(self):
        return self.key

    def set_base_quantity(self, quantity):
        self.base_quantity = quantity
        return None


@pod
class FreeNode:
    next: U32


@pod
class NodeKind(Enum[U64]):
    Uninitialized = None
    Inner = None
    Leaf = None
    Free = None
    LastFree = None


@pod
class Node:
    kind: NodeKind
    node_data: Union[InnerNode, LeafNode, FreeNode, None]

    @classmethod
    def from_bytes_partial(cls, buffer):
        kind, buffer = BYTES_CATALOG.unpack_partial(NodeKind, BytesIO(buffer))
        if kind == NodeKind.Inner:
            node_data, _ = BYTES_CATALOG.unpack_partial(InnerNode, BytesIO(buffer))
        elif kind == NodeKind.Leaf:
            node_data, _ = BYTES_CATALOG.unpack_partial(LeafNode, BytesIO(buffer))
        elif (kind == NodeKind.Free) or (kind == NodeKind.LastFree):
            node_data, _ = BYTES_CATALOG.unpack_partial(FreeNode, BytesIO(buffer))
        else:
            node_data = None
        max_size = max(
            BYTES_CATALOG.calcmaxsize_for_type(InnerNode),
            BYTES_CATALOG.calcmaxsize_for_type(LeafNode),
            BYTES_CATALOG.calcmaxsize_for_type(FreeNode),
        )
        return Node(kind, node_data), buffer[max_size:]


@pod
class SlabHeader:
    account_tag: AccountTag
    bump_index: U64
    free_list_len: U64
    free_list_head: U32
    callback_memory_offset: U64
    callback_free_list_len: U64
    callback_free_list_head: U64
    callback_bump_index: U64
    root_node: U32
    leaf_count: U64
    market_address: PublicKey


@pod
class Slab:
    header: SlabHeader
    # register: Option[OrderSummary]
    buffer: bytes = field(repr=False)

    def __getitem__(self, idx):
        # if idx > self.header.leaf_count:
        #     raise ValueError("Index out of bound")

        start = PADDED_SLAB_HEADER_LEN + idx * SLOT_SIZE
        end = start + SLOT_SIZE
        chunk = self.buffer[start:end]
        return Node.from_bytes_partial(chunk)[0]

    @property
    def root(self):
        if self.header.leaf_count == 0:
            return None
        else:
            return self.header.root_node

    def get_node(self, key):
        start = PADDED_SLAB_HEADER_LEN + key * SLOT_SIZE
        end = start + NODE_SIZE + NODE_TAG_SIZE
        buffer = self.buffer[start:end]
        tag = struct.unpack("<Q", buffer[:8])[0]
        if tag == 1:
            return InnerNode.from_bytes(buffer[8:])
        if tag == 2:
            return LeafNode.from_bytes(buffer[8:])
        if tag == 3 or tag == 4:
            return FreeNode.from_bytes(buffer[8:])

    def find_min_max(self, find_max: bool):
        root = self.root
        if root is None:
            return None
        else:
            while 1:
                data_buffer = self.get_node(root)
                tag, data_buffer = BYTES_CATALOG.unpack_partial(
                    NodeKind, BytesIO(data_buffer)
                )
                if tag == NodeKind.Inner:
                    root_contents, _ = BYTES_CATALOG.unpack_partial(
                        InnerNode, BytesIO(data_buffer)
                    )
                    idx = 1 if find_max else 0
                    root = root_contents.children[idx]  # returns key
                elif (tag == NodeKind.Free) or (tag == NodeKind.LastFree):
                    root_contents, _ = BYTES_CATALOG.unpack_partial(
                        FreeNode, BytesIO(data_buffer)
                    )
                    return root_contents
                elif tag == NodeKind.Leaf:
                    root_contents, _ = BYTES_CATALOG.unpack_partial(
                        LeafNode, BytesIO(data_buffer)
                    )
                    return root_contents
                else:
                    return None

    def inorder_traversal(self, root, side: Side):
        node = self.get_node(root)
        res = []
        if isinstance(node, InnerNode):
            first = 0 if side == Side.BID else 1
            second = 1 - first

            first_child = node.children[first]
            second_child = node.children[second]
            res.extend(self.inorder_traversal(first_child, side))
            res.extend(self.inorder_traversal(second_child, side))

        elif isinstance(node, LeafNode):
            res.extend(
                [
                    (
                        node.price,
                        node.order_id,
                        node.base_quantity,
                    )
                ]
            )

        else:
            res.extend([(None, None, None)])

        return res

    def order_bookify(self, side: Side, group=False):
        root = self.root
        order_book = {}
        if root is None:
            return order_book
        else:
            orders = self.inorder_traversal(root, side)
            for ord in orders:
                price = ord[0]
                oid = ord[1]
                order_size = ord[-1]
                if (price, oid) in order_book:
                    order_book[(price, oid)][0] += order_size
                else:
                    order_book[(price, oid)] = [order_size]
        df = (
            pd.DataFrame(order_book)
            .T.reset_index()
            .rename(columns={"level_0": "Price", "level_1": "Oid", 0: "Qty"})
            .sort_values("Oid", ascending=False)
            .sort_values("Price", ascending=False)
            .reset_index(drop=True)
        )[["Oid", "Price", "Qty"]]
        if group:
            return df.groupby("Price").agg({"Qty": ["sum", "count"]})
        return df

    # key: U128
    # callback_info_pt: U64
    # base_quantity: U64

    @classmethod
    def from_bytes_partial(cls, buffer):
        header, _ = BYTES_CATALOG.unpack_partial(SlabHeader, BytesIO(buffer))

        return Slab(header, buffer), b""

    @classmethod
    def to_bytes_io(cls, obj, buffer):
        raise NotImplementedError



from solana.publickey import PublicKey
from solana.transaction import AccountMeta, TransactionInstruction
from podite import I64, U64, pod

from .common import InstructionCode
from dexteritysdk.program_ids import ORACLE_PROGRAM_ID


@pod
class Params:
    instr: InstructionCode
    slot: U64
    epoch_start_timestamp: I64
    epoch: U64
    leader_schedule_epoch: U64
    unix_timestamp: I64


def update_clock_ix(
    clock: PublicKey,
    slot: int,
    epoch_start_timestamp: int,
    epoch: int,
    leader_schedule_epoch: int,
    unix_timestamp: int,
):
    keys = [
        AccountMeta(pubkey=clock, is_signer=False, is_writable=True),
    ]
    params = Params(
        instr=InstructionCode.UPDATE_CLOCK,
        slot=slot,
        epoch_start_timestamp=epoch_start_timestamp,
        epoch=epoch,
        leader_schedule_epoch=leader_schedule_epoch,
        unix_timestamp=unix_timestamp,
    )
    return TransactionInstruction(
        keys=keys, program_id=ORACLE_PROGRAM_ID, data=params.to_bytes()
    )



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.action_status import ActionStatus
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.codegen.dex.types.health_status import HealthStatus
from dexteritysdk.codegen.dex.types.social_loss import SocialLoss
from podite import (
    FixedLenArray,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(LiquidationInfo)]: DON'T MODIFY
@pod
class LiquidationInfo:
    health: "HealthStatus"
    action: "ActionStatus"
    total_social_loss: "Fractional"
    liquidation_price: "Fractional"
    social_losses: FixedLenArray["SocialLoss", 16]
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from podite import (
    AutoTagType,
    Enum,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(UtilError)]: DON'T MODIFY
@pod
class UtilError(Enum[AutoTagType]):
    ACCOUNT_ALREADY_INITIALIZED = None
    ACCOUNT_UNINITIALIZED = None
    DUPLICATE_PRODUCT_KEY = None
    PUBLIC_KEY_MISMATCH = None
    ASSERTION_ERROR = None
    INVALID_MINT_AUTHORITY = None
    INCORRECT_OWNER = None
    PUBLIC_KEYS_SHOULD_BE_UNIQUE = None
    NOT_RENT_EXEMPT = None
    NUMERICAL_OVERFLOW = None
    ROUND_ERROR = None
    DIVISIONBY_ZERO = None
    INVALID_RETURN_VALUE = None
    SQRT_ROOT_ERROR = None
    ZERO_PRICE_ERROR = None
    ZERO_QUANTITY_ERROR = None
    SERIALIZE_ERROR = None
    DESERIALIZE_ERROR = None
    INVALID_BITSET_INDEX = None
    # LOCK-END

    @classmethod
    def _to_bytes_partial(cls, buffer, obj, **kwargs):
        # to modify packing, change this method
        return super()._to_bytes_partial(buffer, obj, **kwargs)

    @classmethod
    def _from_bytes_partial(cls, buffer, **kwargs):
        # to modify unpacking, change this method
        return super()._from_bytes_partial(buffer, **kwargs)

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from podite import (
    I64,
    U64,
    pod,
)

# LOCK-END

import numpy as np
from typing import Tuple, Union
from decimal import *


MAX_PRECISION = 10

MAX_FRACTIONAL_M = 2 ** 63 - 1
MAX_FRACTIONAL_EXP = 0

MIN_FRACTIONAL_M = -(2 ** 63)
MIN_FRACTIONAL_EXP = 0


# LOCK-BEGIN[class(Fractional)]: DON'T MODIFY
@pod
class Fractional:
    m: I64
    exp: U64
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)

    @classmethod
    def into(cls, x: Union["Fractional", float, int], default_decimals: int):
        if isinstance(x, cls):
            return x
        elif isinstance(x, float):
            return cls(int(x * (10 ** default_decimals)), default_decimals)
        elif isinstance(x, int):
            return cls(x, 0)
        else:
            raise TypeError(f"Expected Fractional, float or int, got {type(x)}")

    @classmethod
    def to_decimal(cls, num):
        num_dec = Decimal(str(float(num)))
        exp = -1 * num_dec.as_tuple().exponent
        num_int = I64(num * (10 ** exp))

        return cls(num_int, exp)

    @property
    def value(self):
        exp = -1 * self.exp
        return self.m * (10 ** exp)

    def __radd__(self, other):
        return Fractional.to_decimal(other) + self

    def __add__(self, other):
        if not isinstance(other, Fractional):
            other = Fractional.to_decimal(other)

        exp = max(self.exp, other.exp)
        m = self.m * 10 ** (exp - self.exp) + other.m * 10 ** (exp - other.exp)
        return Fractional(m, exp).simplify()

    def __rmul__(self, other):
        return Fractional.to_decimal(other) * self

    def __mul__(self, other):
        if not isinstance(other, Fractional):
            other = Fractional.to_decimal(other)

        m = self.m * other.m
        exp = self.exp + other.exp
        return Fractional(m, exp).simplify()

    def __div__(self, other):
        if not isinstance(other, Fractional):
            other = Fractional.to_decimal(other)
        return Fractional.to_decimal(round(self.value / other.value, 6))

    def __rdiv__(self, other):
        if not isinstance(other, Fractional):
            other = Fractional.to_decimal(other)
        return Fractional.to_decimal(round(self.value / other.value, 6))

    def round_sf(self, digits):
        return Fractional.to_decimal(round(self.value, digits))

    def sqrt(self):
        return Fractional.to_decimal(round(np.sqrt(self.value), 2))

    def simplify(self):
        m = self.m
        exp = self.exp

        if m == 0:
            return Fractional(0, 0)

        while m % 10 == 0:
            m //= 10
            exp -= 1
        return Fractional(m, exp)

    def __repr__(self):
        if self.m == MAX_FRACTIONAL_M and self.exp == MAX_FRACTIONAL_EXP:
            return "Inf"
        if self.m == MIN_FRACTIONAL_M and self.exp == MIN_FRACTIONAL_EXP:
            return "-Inf"
        return str(round(self.value, MAX_PRECISION))


from .common import AccountTag

from .clock import (
    Clock,
)
from .oracle_price import (
    OraclePrice,
)


def account_parser(data):
    tag = AccountTag.from_bytes(data[:1], byteorder="little")
    if tag == AccountTag.UNINITIALIZED:
        return None
    elif tag == AccountTag.ORACLE_PRICE:
        return OraclePrice.from_bytes(data)
    raise ValueError()


__all__ = [
    "Clock",
    "AccountTag",
    "OraclePrice",
    "account_parser",
]



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.solmate.dtypes import Usize
from podite import U64

# LOCK-END


# LOCK-BEGIN[constants]: DON'T MODIFY
NAME_LEN: Usize = 16
MAX_OUTRIGHTS: Usize = 128
MAX_PRODUCTS: Usize = 256
HEALTH_BUFFER_LEN: Usize = 32
MAX_TRADER_POSITIONS: Usize = 16
MAX_OPEN_ORDERS_PER_POSITION: U64 = 256
MAX_OPEN_ORDERS: Usize = 1024
ANCHOR_DISCRIMINANT_LEN: Usize = 8
SENTINEL: Usize = 0
CALLBACK_INFO_LEN: U64 = 40
CALLBACK_ID_LEN: U64 = 32
MAX_COMBOS: Usize = 128
MAX_LEGS: Usize = 4
SLOTS_1_MIN: U64 = 150
SLOTS_5_MIN: U64 = 750
SLOTS_15_MIN: U64 = 2250
SLOTS_60_MIN: U64 = 9000
# LOCK-END



import base64
import json
from contextlib import contextmanager
from functools import wraps
from hashlib import sha256
from typing import Optional, Dict
from typing import Union, Tuple, Iterable

import base58
from solana.keypair import Keypair
from solana.publickey import PublicKey
from solana.rpc.api import Client
from solana.rpc.commitment import Confirmed
from solana.rpc.types import TxOpts
from solana.transaction import AccountMeta, Transaction, TransactionInstruction

DEX_LOG_PREFIX = "Program log: dex-log "
ERROR_PREFIX = "Program log: Error: "

def calc_rent(space, client=None):
    if client is None:
        client = Context.get_global_client()
    return client.get_minimum_balance_for_rent_exemption(space)["result"]


class Context:
    client: Optional[Client] = None
    parser: Optional["AccountParser"] = None
    signers: Dict[bytes, Tuple[Keypair, str]] = {}
    fee_payer: Optional[Keypair] = None
    raise_on_error = False

    @staticmethod
    def init_globals(
            fee_payer: Keypair,
            client: Client,
            signers: Iterable[Tuple[Keypair, str]],
            parser: Optional["AccountParser"] = None,
            raise_on_error=False,
    ):
        Context.fee_payer = fee_payer
        Context.client = client
        Context.parser = parser
        Context.signers = {}
        Context.raise_on_error = raise_on_error
        Context.add_signers(*signers)

    @staticmethod
    def get_global_client():
        return Context.client

    @staticmethod
    def set_global_client(client):
        Context.client = client

    @staticmethod
    def get_global_parser():
        return Context.parser

    @staticmethod
    def set_global_parser(parser):
        Context.parser = parser

    @staticmethod
    def get_signers():
        return Context.signers

    # todo: rename to add signers
    @staticmethod
    def add_signers(*signers: Tuple[Keypair, str], verify=True):
        for (signer, name) in signers:
            if not isinstance(signer, Keypair) or not isinstance(name, str):
                raise ValueError(f"signers must be a list iterable of (Keypair, str) tuples. Found: {signer, name}")
            if bytes(signer.public_key) not in Context.signers:
                Context.signers[bytes(signer.public_key)] = (signer, name)
        if verify:
            names = set()
            for (_, name) in Context.signers.values():
                if name in names:
                    raise ValueError("Each signer name must be unique")
                names.add(name)

    @staticmethod
    def get_global_fee_payer():
        return Context.fee_payer

    @staticmethod
    def set_global_fee_payer(fee_payer: Keypair):
        Context.fee_payer = fee_payer

    @staticmethod
    def get_raise_on_error():
        return Context.raise_on_error

    @staticmethod
    def set_raise_on_error(raise_on_error: bool):
        Context.raise_on_error = raise_on_error


@contextmanager
def global_fee_payer(fee_payer):
    old_fee_payer = Context.get_global_fee_payer()
    Context.set_global_fee_payer(fee_payer)
    yield
    Context.set_global_fee_payer(old_fee_payer)


class AccountParser:
    _parsers: Dict[bytes, callable]  # key: program_id

    def __init__(self):
        self._parsers = dict()

    def register_parser(self, program_id, parser):
        self._parsers[bytes(program_id)] = parser

    def register_parser_from_account_enum(self, program_id: PublicKey, accounts_enum):
        def parser(info):
            return accounts_enum.from_bytes(info).field

        self.register_parser(program_id, parser)

    def parse(self, info):
        owner = info["result"]["value"]["owner"]
        data = info["result"]["value"]["data"][0]

        try:
            parser = self._parsers[base58.b58decode(owner)]
        except Exception as e:
            raise ValueError(f"Failed to find parser corresponding to account owner. Owner={owner}",
                             [PublicKey(p).to_base58() for p in self._parsers.keys()])
        return parser(base64.b64decode(data))


class TransactionDetails:
    def __init__(self, content, cluster="devnet"):
        self.content = content
        self.cluster = cluster

    @property
    def account_keys(self):
        accounts = self.content["result"]["transaction"]["message"]["accountKeys"]
        return [PublicKey(account) for account in accounts]

    @property
    def signatures(self):
        return self.content["result"]["transaction"]["signatures"]

    @property
    def tx_string(self):
        return self.signatures[0]

    @property
    def instructions(self):
        account_keys = self.account_keys
        raw_instructions = self.content["result"]["transaction"]["message"][
            "instructions"
        ]
        return [
            dict(
                accounts=[account_keys[int(i)] for i in instr["accounts"]],
                data=instr["data"],
            )
            for instr in raw_instructions
        ]

    @property
    def log_messages(self):
        return self.content["result"]["meta"]["logMessages"]

    @property
    def emitted_logs(self):
        result = dict()
        for msg in self.log_messages:
            if msg.startswith(DEX_LOG_PREFIX):
                key, val = msg[len(DEX_LOG_PREFIX):].split(" ")
                result[key] = base64.b64decode(val)

        return result

    @property
    def error_from_log(self):
        for msg in self.log_messages:
            if msg.startswith(ERROR_PREFIX):
                return msg[len(ERROR_PREFIX):]
        return None

    @property
    def error(self):
        return self.content["result"]["meta"]["err"]

    def __str__(self) -> str:
        return f"TransactionDetails({self.tx_string})"

    def __repr__(self) -> str:
        return str(self)


class AccountDetails:
    def __init__(self, public_key, content, cluster="devnet"):
        self.public_key = public_key
        self.content = content
        self.cluster = cluster

    def __str__(self) -> str:
        return f"AccountDetails({self.public_key})"

    def __repr__(self) -> str:
        return str(self)

    @property
    def data(self):
        value = self.content["result"]["value"]
        if not value:
            return None

        return value["data"]

    @property
    def data_obj(self):
        parser = Context.get_global_parser()
        return parser.parse(self.content)


def fetch_transaction_details(addr, client=None):
    if client is None:
        client = Context.get_global_client()

    content = client.get_confirmed_transaction(
        addr,
        {
            "encoding": "json",
            "commitment": "confirmed",
        },
    )
    return TransactionDetails(content)


def fetch_account_details(addr, client=None):
    if client is None:
        client = Context.get_global_client()

    content = client.get_account_info(addr, commitment=Confirmed)
    return AccountDetails(addr, content)


def explore(addr):
    if isinstance(addr, str):
        if len(base58.b58decode(addr)) == 64:
            kind = "tx"
        else:
            kind = "acc"
            addr = PublicKey(addr)
    elif isinstance(addr, PublicKey):
        kind = "acc"
    else:
        raise ValueError()

    if kind == "tx":
        return fetch_transaction_details(addr)
    else:
        return fetch_account_details(addr)


def send_instructions(
        *ixs: TransactionInstruction,
        **kwargs
) -> Union[TransactionDetails, AccountDetails]:
    return send_transaction(Transaction().add(*ixs), **kwargs)


def send_transaction(
        tx,
        *signers: Keypair,
        opts=TxOpts(
            skip_preflight=True,
            skip_confirmation=False,
            preflight_commitment=Confirmed,
        ),
        recent_blockhash=None,
        client=None,
        raise_on_error=None,
) -> Union[TransactionDetails, AccountDetails]:
    if fee_payer := Context.get_global_fee_payer():
        tx = Transaction(fee_payer=fee_payer.public_key).add(tx)

    raise_on_error = raise_on_error if raise_on_error is not None else Context.get_raise_on_error()

    if len(signers) == 0:
        signers = Context.get_signers()
    else:
        signers = {bytes(signer.public_key): (signer, f"arg  {i}") for i, signer in enumerate(signers)}

    if client is None:
        client = Context.get_global_client()

    # filtering private keys to only contain the relevant ones
    # otherwise, there will be a problem with fee_payer
    signers_public_keys = []
    if tx.fee_payer:
        signers_public_keys.append(tx.fee_payer)
    for ix in tx.instructions:
        for i, meta in enumerate(ix.keys):
            if not isinstance(meta, AccountMeta):
                print(f'{i} is {meta}')
            if meta.is_signer and meta.pubkey not in signers_public_keys:
                signers_public_keys.append(meta.pubkey)

    signer_keypairs = []
    for pk in signers_public_keys:
        if bytes(pk) not in signers:
            names = [(name, PublicKey(p).to_base58()) for p, (_, name) in signers.items()]
            raise ValueError(f"Required signer PublicKey not in list of Keypairs. Have {names}, want: {pk}")
        signer_keypairs.append(signers[bytes(pk)][0])

    result = client.send_transaction(
        tx, *signer_keypairs, opts=opts, recent_blockhash=recent_blockhash
    )
    parsed_result = explore(result["result"])
    if parsed_result.error and raise_on_error:
        err_str = json.dumps(parsed_result.error)
        log_str = "\n".join(parsed_result.log_messages)
        raise ValueError(
            f"Transaction returned error:\n{err_str}, \nLog messages:\n{log_str}")

    return parsed_result


def actionify(func=None, /, post_process=lambda resp: (None, resp), raise_error=False):
    assert not raise_error, "Raise_error is not implemented"

    def _actionify(make):
        @wraps(make)
        def send(*args, **kwargs):
            tx = make(*args, **kwargs)
            if tx is None:
                return post_process(None)
            if isinstance(tx, TransactionInstruction):
                tx = Transaction().add(tx)

            opts = TxOpts(
                skip_preflight=True,
                skip_confirmation=False,
                preflight_commitment=Confirmed,
            )
            response = send_transaction(
                tx,
                opts=opts,
            )
            return post_process(response)

        send.make = make
        return send

    if func is None:
        return _actionify
    return _actionify(func)


def sighash(ix_name: str) -> bytes:
    """Not technically sighash, since we don't include the arguments.
    (Because Rust doesn't allow function overloading.)
    Args:
        ix_name: The instruction name.
    Returns:
        The sighash bytes.
    """
    formatted_str = f"global:{ix_name}"
    return sha256(formatted_str.encode()).digest()[:8]


def sighash_int(ix_name: str) -> int:
    return int.from_bytes(sighash(ix_name), byteorder="little")



# LOCK-BEGIN[imports]: DON'T MODIFY
from .account_tag import AccountTag
from .action_status import ActionStatus
from .bitset import Bitset
from .call_back_info import CallBackInfo
from .cancel_order_params import CancelOrderParams
from .capital_limits import CapitalLimits
from .capital_limits_params import CapitalLimitsParams
from .clear_expired_orderbook_params import ClearExpiredOrderbookParams
from .combo import Combo
from .consume_orderbook_events_params import ConsumeOrderbookEventsParams
from .deposit_funds_params import DepositFundsParams
from .dex_error import DexError
from .fractional import Fractional
from .health_info import HealthInfo
from .health_result import HealthResult
from .health_status import HealthStatus
from .initialize_combo_params import InitializeComboParams
from .initialize_market_product_group_params import InitializeMarketProductGroupParams
from .initialize_market_product_params import InitializeMarketProductParams
from .initialize_print_trade_params import InitializePrintTradeParams
from .leg import Leg
from .liquidation_info import LiquidationInfo
from .market_product_group import MarketProductGroup
from .new_order_params import NewOrderParams
from .open_orders import OpenOrders
from .open_orders_metadata import OpenOrdersMetadata
from .open_orders_node import OpenOrdersNode
from .operation_type import OperationType
from .order_info import OrderInfo
from .order_type import OrderType
from .outright import Outright
from .params import Params
from .price_ewma import PriceEwma
from .print_trade import PrintTrade
from .product import Product
from .product_array import ProductArray
from .product_metadata import ProductMetadata
from .product_status import ProductStatus
from .risk_output_register import RiskOutputRegister
from .sign_print_trade_params import SignPrintTradeParams
from .social_loss import SocialLoss
from .trader_fee_params import TraderFeeParams
from .trader_fees import TraderFees
from .trader_position import TraderPosition
from .trader_risk_group import TraderRiskGroup
from .update_market_product_group_params import UpdateMarketProductGroupParams
from .update_product_funding_params import UpdateProductFundingParams
from .util_error import UtilError
from .withdraw_funds_params import WithdrawFundsParams

# LOCK-END



import struct

from solana.publickey import PublicKey

from dexteritysdk.program_ids import INSTRUMENTS_PROGRAM_ID

MAX_DATES = 32


def get_fixed_income_key(
    market_product_group: PublicKey,
    initialization_time: int,
    coupon_rate: int,
    maturity_date: int,
):
    fixed_income_metadata, bump_seed = PublicKey.find_program_address(
        seeds=[
            b"fixed_income",
            bytes(market_product_group),
            struct.pack("<Q", initialization_time),
            struct.pack("<q", coupon_rate),
            struct.pack("<Q", maturity_date),  # Todo: Add bump seed
        ],
        program_id=INSTRUMENTS_PROGRAM_ID,
    )

    return fixed_income_metadata, bump_seed



from solana.publickey import PublicKey
from solana.transaction import AccountMeta, TransactionInstruction
from solana.system_program import SYS_PROGRAM_ID
from podite import I64, U64, pod

from .common import InstructionCode
from dexteritysdk.program_ids import ORACLE_PROGRAM_ID


@pod
class Params:
    instr: InstructionCode
    price: I64
    decimals: U64


def initialize_oracle_ix(
    oracle_price: PublicKey, update_authority: PublicKey, price: int, decimals: int = 0
):
    keys = [
        AccountMeta(pubkey=oracle_price, is_signer=False, is_writable=True),
        AccountMeta(pubkey=update_authority, is_signer=True, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    params = Params(
        instr=InstructionCode.INITIALIZE_ORACLE, price=price, decimals=decimals
    )
    return TransactionInstruction(
        keys=keys, program_id=ORACLE_PROGRAM_ID, data=params.to_bytes()
    )



import os

import solana.system_program
import solana.sysvar
import spl.token.constants
from solana.publickey import PublicKey
# from solmate import set_pid_by_protocol_name

from dexteritysdk.scripts import extract_program_ids

try:
    programs = extract_program_ids.get_program_to_id()
except:
    programs = {}

RENT_PROGRAM_ID = solana.sysvar.SYSVAR_RENT_PUBKEY
CLOCK_PROGRAM_ID = solana.sysvar.SYSVAR_CLOCK_PUBKEY
SYSTEM_PROGRAM_ID = solana.system_program.SYS_PROGRAM_ID
SPL_TOKEN_PROGRAM_ID = spl.token.constants.TOKEN_PROGRAM_ID

DEX_PROGRAM_ID = PublicKey(
    os.environ.get("DEX", programs.get("dex", ""))
)
INSTRUMENTS_PROGRAM_ID = PublicKey(
    os.environ.get("INSTRUMENTS", programs.get("instruments", ""))
)
ORACLE_PROGRAM_ID = PublicKey(
    os.environ.get("DUMMY_ORACLE", programs.get("dummy_oracle", ""))
)
# NOOP_RISK_ENGINE_PROGRAM_ID = PublicKey(
#     os.environ.get("NOOP_RISK_ENGINE", programs.get("noop_risk_engine", ""))
# )
# SIMPLE_RISK_ENGINE_PROGRAM_ID = PublicKey(
#     os.environ.get("SIMPLE_RISK_ENGINE", "4HzqJ5fXcE5W1YTLa8m6M3cdSuGGjuBmDy9zNSTSnkJY")  # todo remove
# )
# ALPHA_RISK_ENGINE_PROGRAM_ID = PublicKey(
#     os.environ.get("ALPHA_RISK_ENGINE", programs.get("alpha_risk_engine", ""))
# )

# this is to get things going as this is imported in a few places but seems unused
ALPHA_RISK_ENGINE_PROGRAM_ID = PublicKey(1)

RISK_ENGINE_PROGRAM_ID = PublicKey(
    os.environ.get("RISK_ENGINE", "")
)
AOB_PROGRAM_ID = PublicKey(
    os.environ.get("AGNOSTIC_ORDERBOOK",
                   programs.get("agnostic_orderbook", "aaobKniTtDGvCZces7GH5UReLYP671bBkB96ahr9x3e"))
)
CONSTANT_FEES_MODEL_PROGRAM_ID = PublicKey(
    os.environ.get("CONSTANT_FEES", programs.get("constant_fees", ""))
)

# todo: make this ~better~ -> work...
# set_pid_by_protocol_name("risk", ALPHA_RISK_ENGINE_PROGRAM_ID)
# set_pid_by_protocol_name("dex", DEX_PROGRAM_ID)
# set_pid_by_protocol_name("instruments", INSTRUMENTS_PROGRAM_ID)



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, List, NamedTuple, Optional

from construct import ListContainer

from .._layouts.slab import SLAB_LAYOUT, NodeType


class SlabHeader(NamedTuple):
    tag: int
    bump_index: int
    free_list_length: int
    free_list_root: int
    root: int
    leaf_count: int


# Used as dummy value for SlabNode#next.
NONE_NEXT = -1


# UninitializedNode, FreeNode and LastFreeNode all maps to this class.
@dataclass(frozen=True)
class SlabNode:
    is_initialized: bool
    next: int


@dataclass(frozen=True)
class SlabLeafNode(SlabNode):
    key: int
    quantity: int
    client_order_id: int


@dataclass(frozen=True)
class SlabInnerNode(SlabNode):
    prefix_len: int
    key: int
    children: List[int]


class Slab:
    def __init__(self, header: SlabHeader, nodes: List[SlabNode]):
        self._header: SlabHeader = header
        self._nodes: List[SlabNode] = nodes

    @staticmethod
    def __build(nodes: ListContainer) -> List[SlabNode]:
        res: List[SlabNode] = []
        for construct_node in nodes:
            node_type = construct_node.tag
            node = construct_node.node
            if node_type == NodeType.UNINITIALIZED:
                res.append(SlabNode(is_initialized=False, next=-1))
            elif node_type == NodeType.LEAF_NODE:
                res.append(
                    SlabLeafNode(
                        key=int.from_bytes(node.key, "little"),
                        quantity=node.quantity,
                        client_order_id=node.callback,
                        is_initialized=True,
                        next=NONE_NEXT,
                    )
                )
            elif node_type == NodeType.INNER_NODE:
                res.append(
                    SlabInnerNode(
                        prefix_len=node.prefix_len,
                        key=int.from_bytes(node.key, "little"),
                        children=node.children,
                        is_initialized=True,
                        next=NONE_NEXT,
                    )
                )
            elif node_type == NodeType.FREE_NODE:
                res.append(SlabNode(is_initialized=True, next=node.next))
            elif node_type == NodeType.LAST_FREE_NODE:
                res.append(SlabNode(is_initialized=True, next=NONE_NEXT))
            else:
                raise RuntimeError("Unrecognized node type" + node.tag)
        return res

    @staticmethod
    def from_bytes(buffer: bytes) -> Slab:
        parsed_slab = SLAB_LAYOUT.parse(buffer)
        header = parsed_slab.header
        nodes = parsed_slab.nodes
        return Slab(
            SlabHeader(
                tag=header.tag,
                bump_index=header.bump_index,
                free_list_length=header.free_list_length,
                free_list_root=header.free_list_head,
                root=header.root,
                leaf_count=header.leaf_count,
            ),
            Slab.__build(nodes),
        )

    def get(self, search_key: int) -> Optional[SlabLeafNode]:
        if self._header.leaf_count == 0:
            return None
        index: int = self._header.root
        while True:
            node: SlabNode = self._nodes[index]
            if isinstance(node, SlabLeafNode):  # pylint: disable=no-else-return
                return node if node.key == search_key else None
            elif isinstance(node, SlabInnerNode):
                if (node.key ^ search_key) >> (128 - node.prefix_len) != 0:
                    return None
                # Check if the n-th bit (start from the least significant, i.e. rightmost) of the key is set
                index = node.children[(search_key >> (128 - node.prefix_len - 1)) & 1]
            else:
                raise RuntimeError("Should not go here! Node type not recognize.")

    def __iter__(self) -> Iterable[SlabLeafNode]:
        return self.items(False)

    def items(self, descending=False) -> Iterable[SlabLeafNode]:
        """Depth first traversal of the Binary Tree.
        Parameter descending decides if the price should descending or not.
        """
        if self._header.leaf_count == 0:
            return
        stack = [self._header.root]
        while stack:
            index = stack.pop()
            node: SlabNode = self._nodes[index]
            if isinstance(node, SlabLeafNode):
                yield node
            elif isinstance(node, SlabInnerNode):
                if descending:
                    stack.append(node.children[0])
                    stack.append(node.children[1])
                else:
                    stack.append(node.children[1])
                    stack.append(node.children[0])
            else:
                raise RuntimeError("Neither of leaf node or tree node!")



from hashlib import sha256

from podite import U64


def sighash(namespace: str, name: str) -> int:
    preimage = f"{namespace}:{name}".encode()
    hash_bytes = sha256(preimage).digest()

    return U64.from_bytes(hash_bytes[:8])



# LOCK-BEGIN[imports]: DON'T MODIFY
from .instruction_tag import InstructionTag
from dataclasses import dataclass
from dexteritysdk.codegen.dex.types import SignPrintTradeParams
from dexteritysdk.solmate.utils import to_account_meta
from io import BytesIO
from podite import BYTES_CATALOG
from solana.publickey import PublicKey
from solana.transaction import (
    AccountMeta,
    TransactionInstruction,
)
from typing import (
    List,
    Optional,
    Union,
)

# LOCK-END


# LOCK-BEGIN[ix_cls(sign_print_trade)]: DON'T MODIFY
@dataclass
class SignPrintTradeIx:
    program_id: PublicKey

    # account metas
    user: AccountMeta
    creator: AccountMeta
    counterparty: AccountMeta
    operator: AccountMeta
    market_product_group: AccountMeta
    product: AccountMeta
    print_trade: AccountMeta
    system_program: AccountMeta
    fee_model_program: AccountMeta
    fee_model_configuration_acct: AccountMeta
    fee_output_register: AccountMeta
    risk_engine_program: AccountMeta
    risk_model_configuration_acct: AccountMeta
    risk_output_register: AccountMeta
    risk_and_fee_signer: AccountMeta
    creator_trader_fee_state_acct: AccountMeta
    creator_trader_risk_state_acct: AccountMeta
    counterparty_trader_fee_state_acct: AccountMeta
    counterparty_trader_risk_state_acct: AccountMeta
    remaining_accounts: Optional[List[AccountMeta]]

    # data fields
    params: SignPrintTradeParams

    def to_instruction(self):
        keys = []
        keys.append(self.user)
        keys.append(self.creator)
        keys.append(self.counterparty)
        keys.append(self.operator)
        keys.append(self.market_product_group)
        keys.append(self.product)
        keys.append(self.print_trade)
        keys.append(self.system_program)
        keys.append(self.fee_model_program)
        keys.append(self.fee_model_configuration_acct)
        keys.append(self.fee_output_register)
        keys.append(self.risk_engine_program)
        keys.append(self.risk_model_configuration_acct)
        keys.append(self.risk_output_register)
        keys.append(self.risk_and_fee_signer)
        keys.append(self.creator_trader_fee_state_acct)
        keys.append(self.creator_trader_risk_state_acct)
        keys.append(self.counterparty_trader_fee_state_acct)
        keys.append(self.counterparty_trader_risk_state_acct)
        if self.remaining_accounts is not None:
            keys.extend(self.remaining_accounts)

        buffer = BytesIO()
        buffer.write(InstructionTag.to_bytes(InstructionTag.SIGN_PRINT_TRADE))
        buffer.write(BYTES_CATALOG.pack(SignPrintTradeParams, self.params))

        return TransactionInstruction(
            keys=keys,
            program_id=self.program_id,
            data=buffer.getvalue(),
        )

# LOCK-END


# LOCK-BEGIN[ix_fn(sign_print_trade)]: DON'T MODIFY
def sign_print_trade(
    user: Union[str, PublicKey, AccountMeta],
    creator: Union[str, PublicKey, AccountMeta],
    counterparty: Union[str, PublicKey, AccountMeta],
    operator: Union[str, PublicKey, AccountMeta],
    market_product_group: Union[str, PublicKey, AccountMeta],
    product: Union[str, PublicKey, AccountMeta],
    print_trade: Union[str, PublicKey, AccountMeta],
    fee_model_program: Union[str, PublicKey, AccountMeta],
    fee_model_configuration_acct: Union[str, PublicKey, AccountMeta],
    fee_output_register: Union[str, PublicKey, AccountMeta],
    risk_engine_program: Union[str, PublicKey, AccountMeta],
    risk_model_configuration_acct: Union[str, PublicKey, AccountMeta],
    risk_output_register: Union[str, PublicKey, AccountMeta],
    risk_and_fee_signer: Union[str, PublicKey, AccountMeta],
    creator_trader_fee_state_acct: Union[str, PublicKey, AccountMeta],
    creator_trader_risk_state_acct: Union[str, PublicKey, AccountMeta],
    counterparty_trader_fee_state_acct: Union[str, PublicKey, AccountMeta],
    counterparty_trader_risk_state_acct: Union[str, PublicKey, AccountMeta],
    params: SignPrintTradeParams,
    system_program: Union[str, PublicKey, AccountMeta] = PublicKey("11111111111111111111111111111111"),
    remaining_accounts: Optional[List[AccountMeta]] = None,
    program_id: Optional[PublicKey] = None,
):
    if program_id is None:
        program_id = PublicKey("FUfpR31LmcP1VSbz5zDaM7nxnH55iBHkpwusgrnhaFjL")

    if isinstance(user, (str, PublicKey)):
        user = to_account_meta(
            user,
            is_signer=True,
            is_writable=True,
        )
    if isinstance(creator, (str, PublicKey)):
        creator = to_account_meta(
            creator,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(counterparty, (str, PublicKey)):
        counterparty = to_account_meta(
            counterparty,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(operator, (str, PublicKey)):
        operator = to_account_meta(
            operator,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(market_product_group, (str, PublicKey)):
        market_product_group = to_account_meta(
            market_product_group,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(product, (str, PublicKey)):
        product = to_account_meta(
            product,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(print_trade, (str, PublicKey)):
        print_trade = to_account_meta(
            print_trade,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(system_program, (str, PublicKey)):
        system_program = to_account_meta(
            system_program,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(fee_model_program, (str, PublicKey)):
        fee_model_program = to_account_meta(
            fee_model_program,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(fee_model_configuration_acct, (str, PublicKey)):
        fee_model_configuration_acct = to_account_meta(
            fee_model_configuration_acct,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(fee_output_register, (str, PublicKey)):
        fee_output_register = to_account_meta(
            fee_output_register,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(risk_engine_program, (str, PublicKey)):
        risk_engine_program = to_account_meta(
            risk_engine_program,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(risk_model_configuration_acct, (str, PublicKey)):
        risk_model_configuration_acct = to_account_meta(
            risk_model_configuration_acct,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(risk_output_register, (str, PublicKey)):
        risk_output_register = to_account_meta(
            risk_output_register,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(risk_and_fee_signer, (str, PublicKey)):
        risk_and_fee_signer = to_account_meta(
            risk_and_fee_signer,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(creator_trader_fee_state_acct, (str, PublicKey)):
        creator_trader_fee_state_acct = to_account_meta(
            creator_trader_fee_state_acct,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(creator_trader_risk_state_acct, (str, PublicKey)):
        creator_trader_risk_state_acct = to_account_meta(
            creator_trader_risk_state_acct,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(counterparty_trader_fee_state_acct, (str, PublicKey)):
        counterparty_trader_fee_state_acct = to_account_meta(
            counterparty_trader_fee_state_acct,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(counterparty_trader_risk_state_acct, (str, PublicKey)):
        counterparty_trader_risk_state_acct = to_account_meta(
            counterparty_trader_risk_state_acct,
            is_signer=False,
            is_writable=True,
        )

    return SignPrintTradeIx(
        program_id=program_id,
        user=user,
        creator=creator,
        counterparty=counterparty,
        operator=operator,
        market_product_group=market_product_group,
        product=product,
        print_trade=print_trade,
        system_program=system_program,
        fee_model_program=fee_model_program,
        fee_model_configuration_acct=fee_model_configuration_acct,
        fee_output_register=fee_output_register,
        risk_engine_program=risk_engine_program,
        risk_model_configuration_acct=risk_model_configuration_acct,
        risk_output_register=risk_output_register,
        risk_and_fee_signer=risk_and_fee_signer,
        creator_trader_fee_state_acct=creator_trader_fee_state_acct,
        creator_trader_risk_state_acct=creator_trader_risk_state_acct,
        counterparty_trader_fee_state_acct=counterparty_trader_fee_state_acct,
        counterparty_trader_risk_state_acct=counterparty_trader_risk_state_acct,
        remaining_accounts=remaining_accounts,
        params=params,
    ).to_instruction()

# LOCK-END



#!/usr/bin/env python3

# python quoter-bot.py 7 Crypto.BTC/USD AVG4tHmvfGssaDLNddVqbkwwjJH3dPi7h3Yhg7ZqjfaM jarry 10
# python quoter_bot.py BTC_0 Crypto.BTC/USD 9UW8hNAzUhJ9mBScLgcQZYoSLZoPpDqgiFAzF3wP3n5q jarry 1
# python quoter_bot.py BTC_0 Crypto.BTC/USD jjcJHDJ6x7kaTVRrz8GKPYH1y2pUpzRqJHNfDYpHHRy jarry 1
from __future__ import annotations
from dexteritysdk.local_env import *
import asyncio
import signal
import sys
from typing import List, Any
import argparse
import random
import time
import json
import numpy as np

from loguru import logger

from pythclient.pythclient import PythClient  # noqa
from pythclient.ratelimit import RateLimit  # noqa
from pythclient.pythaccounts import PythPriceAccount  # noqa
from pythclient.utils import get_key  # noqa
from solana.keypair import Keypair
from solana.rpc.api import Client
from solana.transaction import Transaction
from solana.rpc.async_api import AsyncClient


from dexteritysdk.program_ids import *
import dexteritysdk.utils.solana as ctx
import dexteritysdk.dex.actions as dex_act
import dexteritysdk.dex.instructions as dex_ixs
from dexteritysdk.utils.solana import explore, AccountParser
from dexteritysdk.dex.state import account_parser as dex_account_parser
from dexteritysdk.dex.state import OpenOrdersNode
from dexteritysdk.utils.aob import account_parser as aob_account_parser
from dexteritysdk.dex.instructions.common import Side

from dexteritysdk.utils import binance_parser

from solana.rpc.types import TxOpts
from solana.rpc.commitment import Confirmed

MAX_MARKET_PRODUCT = 128

logger.enable("pythclient")

RateLimit.configure_default_ratelimit(overall_cps=9, method_cps=3, connection_cps=3)

to_exit = False

parser = AccountParser()
parser.register_parser(DEX_PROGRAM_ID, dex_account_parser)
parser.register_parser(AOB_PROGRAM_ID, aob_account_parser)


ctx.set_global_parser(parser)
url = "https://api.devnet.solana.com"
ctx.set_global_client(Client(url))

curr_directory = os.path.dirname(os.path.abspath(__file__))
cfg_path = os.path.join(curr_directory, "config.json")
if os.path.exists(cfg_path):
    with open(cfg_path) as f:
        cfg = json.load(f)
        mpg_key = cfg["mpg_key"]

        ap = argparse.ArgumentParser()
        ap.add_argument("prod_name")
        ap.add_argument("binance_ticker")
        ap.add_argument("size", nargs="?", const=1, default=1, type=int)
        ap.add_argument("idx", nargs="?", const=1, default=0, type=int)
        ap.add_argument("positions_limit", nargs="?", const=1, default=1, type=int)
        args = ap.parse_args()
        trader = Keypair(cfg["quoters"][args.idx].encode())
        prod_name = args.prod_name
        oracle = cfg["product_oracles"][prod_name]
        size = args.size

else:
    ap = argparse.ArgumentParser()
    ap.add_argument("prod_name")
    ap.add_argument("binance_ticker")
    ap.add_argument("oracle_prod_name")
    ap.add_argument("mpg_pubkey")
    ap.add_argument("seed")
    ap.add_argument("size", nargs="?", const=1, default=1, type=int)
    ap.add_argument("positions_limit", nargs="?", const=1, default=1, type=int)
    args = ap.parse_args()

    mpg_key = args.mpg_pubkey
    prod_name = args.prod_name

    trader = Keypair(
        bytes(args.seed, "utf-8") + b"1" + b"x" * (31 - len(bytes(args.seed, "utf-8")))
    )
    oracle = args.oracle_prod_name
    size = int(args.min_size)


bparser = binance_parser.BinanceParser(args.binance_ticker)


prev_fair = None

ctx.set_signers(trader)


trader_risk_group = dex_act.get_trader_risk_group_addr(trader.public_key, mpg_key)

market_obj = explore(mpg_key).data_obj
risk_engine = market_obj.risk_engine_program


for idx, product in enumerate(market_obj.market_products):
    if product:
        if product.name == prod_name:
            product_obj = product
            product_idx = idx
            break
    else:
        raise
else:
    raise


orderbook_obj = explore(product_obj.orderbook).data_obj

tick_size = product_obj.tick_size.m * 10 ** (-product_obj.tick_size.exp)


def set_to_exit(sig: Any, frame: Any):
    global to_exit
    to_exit = True


signal.signal(signal.SIGINT, set_to_exit)


def new_order(side, price):
    print(f"Send new order: side: {side}, price: {price}")
    return dex_act.new_order(
        trader=trader.public_key,
        market_product_group=mpg_key,
        product_key=product_obj.product_key,
        side=side,
        limit_price=price,
        max_base_qty=size,
        order_type=dex_ixs.OrderType.IMMEDIATE_OR_CANCEL,
    )


async def get_bbo(side, slab_pubkey):
    if side == Side.ASK:
        func = np.min
        bbo = np.inf
    elif side == Side.BID:
        func = np.max
        bbo = -np.inf
    slab = explore(slab_pubkey).data_obj
    counter = 0
    while True:
        if slab[counter].kind.name == "Leaf":
            order_price = slab[counter].node_data.key // 2 ** 64 // 2 ** 32 * tick_size
            bbo = func((bbo, order_price))
        if slab[counter].kind.name == "Uninitialized":
            break
        counter += 1
    print(f"BBO for side: {side}, price: {bbo}")
    return bbo


async def async_send(*args):
    await asyncio.gather(args)


def get_positions_count(trg):
    trg_obj = explore(trg).data_obj
    position_idx = trg_obj.active_positions[product_idx]
    position = trg_obj.trader_positions[position_idx]
    return position.position.value


async def main():
    global to_exit
    while True:
        try:
            best_bid = await get_bbo(Side.BID, orderbook_obj.bids)
            best_ask = await get_bbo(Side.ASK, orderbook_obj.asks)
            # print(f"best bid: {best_bid}")
            # print(f"best ask: {best_ask}")
            book = bparser.get_book()
            bbbo = book.iloc[0]
            bbest_bid = float(bbbo.bid)
            bbest_ask = float(bbbo.ask)
            positions_count = get_positions_count(trader_risk_group)
            if bbest_bid > best_ask and positions_count <= args.positions_limit:
                print("time to buy")
                new_order(Side.BID, best_ask)
            elif bbest_ask < best_bid and positions_count >= -args.positions_limit:
                print("time to sell")
                new_order(Side.ASK, best_bid)
            time.sleep(0.5)
        except Exception as e:
            print(f"Exception occured: {e}")


asyncio.run(main())



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.account_tag import AccountTag
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.codegen.dex.types.open_orders import OpenOrders
from dexteritysdk.codegen.dex.types.trader_position import TraderPosition
from dexteritysdk.solmate.dtypes import UnixTimestamp
from podite import (
    FixedLenArray,
    I32,
    U128,
    U8,
    pod,
)
from solana.publickey import PublicKey

# LOCK-END


# LOCK-BEGIN[class(TraderRiskGroup)]: DON'T MODIFY
@pod
class TraderRiskGroup:
    tag: AccountTag
    market_product_group: PublicKey
    owner: PublicKey
    active_products: FixedLenArray[U8, 128]
    total_deposited: Fractional
    total_withdrawn: Fractional
    cash_balance: Fractional
    pending_cash_balance: Fractional
    pending_fees: Fractional
    valid_until: UnixTimestamp
    maker_fee_bps: I32
    taker_fee_bps: I32
    trader_positions: FixedLenArray[TraderPosition, 16]
    risk_state_account: PublicKey
    fee_state_account: PublicKey
    client_order_id: U128
    open_orders: OpenOrders
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from .close_derivative_account import (
    CloseDerivativeAccountIx,
    close_derivative_account,
)
from .initialize_derivative import (
    InitializeDerivativeIx,
    initialize_derivative,
)
from .instruction_tag import InstructionTag
from .settle_derivative import (
    SettleDerivativeIx,
    settle_derivative,
)

# LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
from .instruction_tag import InstructionTag
from dataclasses import dataclass
from dexteritysdk.solmate.utils import to_account_meta
from io import BytesIO
from podite import BYTES_CATALOG
from solana.publickey import PublicKey
from solana.transaction import (
    AccountMeta,
    TransactionInstruction,
)
from typing import (
    List,
    Optional,
    Union,
)

# LOCK-END


# LOCK-BEGIN[ix_cls(close_derivative_account)]: DON'T MODIFY
@dataclass
class CloseDerivativeAccountIx:
    program_id: PublicKey

    # account metas
    derivative_metadata: AccountMeta
    close_authority: AccountMeta
    destination: AccountMeta
    remaining_accounts: Optional[List[AccountMeta]]

    def to_instruction(self):
        keys = []
        keys.append(self.derivative_metadata)
        keys.append(self.close_authority)
        keys.append(self.destination)
        if self.remaining_accounts is not None:
            keys.extend(self.remaining_accounts)

        buffer = BytesIO()
        buffer.write(InstructionTag.to_bytes(InstructionTag.CLOSE_DERIVATIVE_ACCOUNT))

        return TransactionInstruction(
            keys=keys,
            program_id=self.program_id,
            data=buffer.getvalue(),
        )

# LOCK-END


# LOCK-BEGIN[ix_fn(close_derivative_account)]: DON'T MODIFY
def close_derivative_account(
    derivative_metadata: Union[str, PublicKey, AccountMeta],
    close_authority: Union[str, PublicKey, AccountMeta],
    destination: Union[str, PublicKey, AccountMeta],
    remaining_accounts: Optional[List[AccountMeta]] = None,
    program_id: Optional[PublicKey] = None,
):
    if program_id is None:
        program_id = PublicKey("8981bZYszfz1FrFVx7gcUm61RfawMoAHnURuERRJKdkq")

    if isinstance(derivative_metadata, (str, PublicKey)):
        derivative_metadata = to_account_meta(
            derivative_metadata,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(close_authority, (str, PublicKey)):
        close_authority = to_account_meta(
            close_authority,
            is_signer=True,
            is_writable=False,
        )
    if isinstance(destination, (str, PublicKey)):
        destination = to_account_meta(
            destination,
            is_signer=False,
            is_writable=False,
        )

    return CloseDerivativeAccountIx(
        program_id=program_id,
        derivative_metadata=derivative_metadata,
        close_authority=close_authority,
        destination=destination,
        remaining_accounts=remaining_accounts,
    ).to_instruction()

# LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.utils.aob.state.base import Side
from podite import (
    U8,
    pod,
)
from solana.publickey import PublicKey

# LOCK-END


# LOCK-BEGIN[class(PrintTrade)]: DON'T MODIFY
@pod
class PrintTrade:
    is_initialized: bool
    creator: PublicKey
    counterparty: PublicKey
    market_product_group: PublicKey
    product: PublicKey
    size: Fractional
    price: Fractional
    side: Side
    operator: PublicKey
    operator_creator_fee_proportion: Fractional
    operator_counterparty_fee_proportion: Fractional
    bump: U8
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



import base64
from dataclasses import dataclass
from typing import List, Union, Optional

import base58
from solana.rpc import types
from solana.rpc.commitment import Confirmed
from spl.token import instructions as spl_token_instructions

from ..pyserum.orderbook import OrderBook

from solana.rpc.types import TxOpts
from solana.keypair import Keypair
from solana.publickey import PublicKey
from solana.system_program import SYS_PROGRAM_ID
from solana.transaction import AccountMeta, TransactionInstruction
from solana import system_program

import dexteritysdk.codegen.dex.instructions as dixs
import dexteritysdk.codegen.dex.types as dtys
from dexteritysdk import program_ids as pids, mints
from dexteritysdk.codegen.dex.accounts import Accounts as DexAccounts
from dexteritysdk.codegen.dex.constants import SENTINEL
from dexteritysdk.codegen.dex.types import MarketProductGroup, Product, CancelOrderParams
from dexteritysdk.constant_fees import instructions as fee_ixs
from dexteritysdk.dex import addrs as daddrs
from dexteritysdk.dex.actions import DexOrderSummary
from dexteritysdk.utils import solana as solana_utils
from dexteritysdk.utils.aob import state as aaob_state
from dexteritysdk.codegen.instruments.accounts import Accounts as InstrumentAccounts
from dexteritysdk.utils.solana import Client, Context, AccountParser, explore, fetch_account_details
from dexteritysdk.solmate.utils import to_account_meta


@dataclass
class SDKOrder:
    order_id: int
    product: "SDKProduct"
    side: aaob_state.Side
    price: float
    qty: float

    @staticmethod
    def from_pyserum_orderbook(order_book: OrderBook, product: "SDKProduct", bids: bool) -> List["SDKOrder"]:
        return [
            SDKOrder(
                o.order_id,
                product,
                aaob_state.Side.BID if bids else aaob_state.Side.ASK,
                o.info.price,
                o.info.size
            ) for o in order_book.orders()
        ]


@dataclass
class SDKOrderSummary:
    order_id: int
    product: "SDKProduct"
    side: aaob_state.Side
    price: float
    qty: float
    filled_qty: float
    remaining_qty: float

    @staticmethod
    def from_dex_order_summary(sdk: "SDKContext", order_summary: DexOrderSummary, product: "SDKProduct",
                               side: aaob_state.Side, quantity: Union[dtys.Fractional, float],
                               price: Union[dtys.Fractional, float]) -> "SDKOrderSummary":
        if isinstance(quantity, dtys.Fractional):
            quantity = quantity.value
        if isinstance(price, dtys.Fractional):
            price = price.value
        order_id = None if order_summary.posted_order_id == order_summary.posted_order_id.NONE \
            else order_summary.posted_order_id.field
        remaining_qty = order_summary.total_base_qty_posted / (10 ** sdk.decimals)
        filled_qty = quantity - remaining_qty
        return SDKOrderSummary(
            order_id=order_id,
            product=product,
            side=side,
            price=price,
            qty=quantity,
            filled_qty=filled_qty,
            remaining_qty=remaining_qty
        )

@dataclass
class SDKProduct:
    key: PublicKey
    name: str  # max 16 bytes
    orderbook: PublicKey
    bids: PublicKey
    asks: PublicKey
    market_signer: PublicKey
    event_queue: PublicKey
    stale_product: Product  # last time prod was fetched

    def get_orderbook(self, sdk: "SDKContext") -> "SDKOrderBook":
        return SDKOrderBook.from_product(sdk, self)

    def crank_raw(self, sdk: "SDKContext", trader_and_risk_accounts: List[PublicKey], reward_target: PublicKey):
        trader_and_risk_accounts.sort()
        trader_and_risk_accounts = _dedup(trader_and_risk_accounts)
        ix = dixs.consume_orderbook_events(
            aaob_program=sdk.aaob_program,
            market_product_group=sdk.market_product_group,
            product=self.key,
            market_signer=self.market_signer,
            orderbook=self.orderbook,
            event_queue=self.event_queue,
            reward_target=reward_target,
            fee_model_program=sdk.fee_model_program,
            fee_model_configuration_acct=sdk.fee_model_configuration_acct,
            fee_output_register=sdk.fee_output_register,
            risk_and_fee_signer=daddrs.get_risk_signer(sdk.market_product_group),
            params=dtys.ConsumeOrderbookEventsParams(
                20
            ),
            remaining_accounts=[AccountMeta(pk, False, False) for pk in trader_and_risk_accounts]
        )
        solana_utils.send_instructions(ix)


def _dedup(xs):
    if len(xs) < 2:
        return xs
    i = 0
    for acct in xs[1:]:
        if xs[i] != acct:
            i += 1
            xs[i] = acct
    return xs[:i + 1]


@dataclass
class SDKOrderBook:
    product: SDKProduct
    bids: List[SDKOrder]
    asks: List[SDKOrder]

    @staticmethod
    def from_product(sdk: "SDKContext", product: "SDKProduct") -> "SDKOrderBook":
        bids_data = base64.b64decode(explore(product.bids).data[0])
        asks_data = base64.b64decode(explore(product.asks).data[0])

        return SDKOrderBook(
            product=product,
            bids=SDKOrder.from_pyserum_orderbook(
                order_book=OrderBook.from_bytes(bids_data, sdk.decimals),
                product=product,
                bids=True,
            ),
            asks=SDKOrder.from_pyserum_orderbook(
                order_book=OrderBook.from_bytes(asks_data, sdk.decimals),
                product=product,
                bids=False,
            )
        )


@dataclass
class SDKTrader:
    keypair: Keypair
    account: PublicKey
    wallet: PublicKey
    trader_fee_state_acct: PublicKey
    trader_risk_state_acct: PublicKey  # separate pk vs kp to allow **vars(trader) in ixs
    whitelist_token_wallet: PublicKey
    dex_program: PublicKey = None

    @staticmethod
    def connect(
            sdk: "SDKContext",
            account: PublicKey,
            keypair: Keypair,
            wallet: PublicKey,
            trader_risk_state_acct: PublicKey,
            dex_program: PublicKey = None,
    ) -> "SDKTrader":
        trader_fee_state_acct = daddrs.get_trader_fee_state_acct(account, sdk.market_product_group,
                                                                 sdk.fee_model_program)
        trg: dtys.TraderRiskGroup = explore(account).data_obj
        assert trg.market_product_group == sdk.market_product_group
        assert trg.risk_state_account == trader_risk_state_acct
        assert trg.fee_state_account == trader_fee_state_acct
        assert trg.owner == keypair.public_key

        whitelist_token_wallet = spl_token_instructions.get_associated_token_address(keypair.public_key, mints.WHITELIST_TOKEN_MINT)

        return SDKTrader(
            keypair, account, wallet, trader_fee_state_acct, trader_risk_state_acct, whitelist_token_wallet, dex_program)

    def get_trader_risk_group(self) -> dtys.TraderRiskGroup:
        return fetch_account_details(self.account).data_obj

    def deposit(self, sdk: "SDKContext", qty: Union[float, dtys.Fractional]):
        if not isinstance(qty, dtys.Fractional):
            qty = dtys.Fractional(int(qty * (10 ** sdk.decimals)), sdk.decimals)
        ix = dixs.deposit_funds(
            user=self.keypair.public_key,
            user_token_account=self.wallet,
            trader_risk_group=self.account,
            market_product_group=sdk.market_product_group,
            market_product_group_vault=sdk.market_product_group_vault,
            capital_limits=sdk.capital_limits,
            whitelist_ata_acct=self.whitelist_token_wallet,
            params=dtys.DepositFundsParams(
                quantity=qty,
            ),
            program_id=self.dex_program,
        )
        return sdk.send_instructions(ix)

    def withdraw(self, sdk: "SDKContext", qty: Union[float, dtys.Fractional]):
        if not isinstance(qty, dtys.Fractional):
            qty = dtys.Fractional(int(qty * sdk.decimals), sdk.decimals)
        ix = dixs.withdraw_funds(
            user=self.keypair.public_key,
            user_token_account=self.wallet,
            trader_risk_group=self.account,
            market_product_group=sdk.market_product_group,
            market_product_group_vault=sdk.market_product_group_vault,
            risk_output_register=sdk.risk_output_register,
            risk_engine_program=sdk.risk_engine_program,
            risk_model_configuration_acct=sdk.risk_model_configuration_acct,
            risk_signer=sdk.risk_signer,
            capital_limits=sdk.capital_limits,
            risk_state_account=self.trader_risk_state_acct,
            params=dtys.WithdrawFundsParams(
                quantity=qty,
            ),
            program_id=self.dex_program,
        )
        return sdk.send_instructions(ix)

    def place_order(
            self,
            sdk: "SDKContext",
            product: Union[SDKProduct, PublicKey],
            side: aaob_state.Side,
            size: Union[dtys.Fractional, float],
            price: Union[dtys.Fractional, float],
            self_trade_behavior: aaob_state.SelfTradeBehavior = aaob_state.SelfTradeBehavior.DECREMENT_TAKE,
            order_type: dtys.OrderType = dtys.OrderType.LIMIT
    ) -> SDKOrderSummary:
        ix = self._place_order_ix(sdk, product, side, size, price, self_trade_behavior, order_type,
                                  sdk.additional_risk_accts)
        trans_details = sdk.send_instructions(ix, raise_on_error=False)
        if trans_details.error:
            raise ValueError(trans_details.error_from_log)
        else:
            raw_summary = trans_details.emitted_logs["new-order:order-summary"]
            return SDKOrderSummary.from_dex_order_summary(sdk, DexOrderSummary.from_bytes(raw_summary),
                                                          product, side, size, price)

    def _place_order_ix(
            self,
            sdk: "SDKContext",
            product: Union[SDKProduct, PublicKey],
            side: aaob_state.Side,
            size: Union[dtys.Fractional, float],
            price: Union[dtys.Fractional, float],
            self_trade_behavior: aaob_state.SelfTradeBehavior = aaob_state.SelfTradeBehavior.DECREMENT_TAKE,
            order_type: dtys.OrderType = dtys.OrderType.LIMIT,
            risk_accounts: Optional[List[PublicKey]] = None,
    ):
        remaining_accounts = [to_account_meta(ra, is_signer=False, is_writable=True) for ra in risk_accounts]

        ix = dixs.new_order(
            program_id=self.dex_program,
            user=self.keypair.public_key,
            trader_risk_group=self.account,
            market_product_group=sdk.market_product_group,
            product=product.key,
            aaob_program=sdk.aaob_program,
            orderbook=product.orderbook,
            market_signer=product.market_signer,
            event_queue=product.event_queue,
            bids=product.bids,
            asks=product.asks,
            fee_model_program=sdk.fee_model_program,
            fee_model_configuration_acct=sdk.fee_model_configuration_acct,
            trader_fee_state_acct=self.trader_fee_state_acct,
            fee_output_register=sdk.fee_output_register,
            risk_engine_program=sdk.risk_engine_program,
            risk_model_configuration_acct=sdk.risk_model_configuration_acct,
            risk_output_register=sdk.risk_output_register,
            trader_risk_state_acct=self.trader_risk_state_acct,
            risk_and_fee_signer=sdk.risk_signer,
            params=dtys.NewOrderParams(
                side=side,
                max_base_qty=dtys.Fractional.into(size, product.stale_product.metadata().base_decimals),
                order_type=order_type,
                self_trade_behavior=self_trade_behavior,
                match_limit=10,
                limit_price=dtys.Fractional.into(price, product.stale_product.metadata().base_decimals),
            ),
            remaining_accounts=remaining_accounts,
        )

        return ix

    def cancel(
            self,
            sdk: "SDKContext",
            product: SDKProduct,
            order_id: int,
    ):
        self.cancel_underwater(sdk, product, order_id, self.account)

    def cancel_underwater(
            self,
            sdk: "SDKContext",
            product: SDKProduct,
            order_id: int,
            under_water_trg: PublicKey,
    ):
        ix = self._cancel_ix(sdk, product, order_id, under_water_trg)
        trans_details = sdk.send_instructions(ix, raise_on_error=False)
        if trans_details.error:
            raise ValueError(trans_details.error_from_log)

    def _cancel_ix(
            self,
            sdk: "SDKContext",
            product: SDKProduct,
            order_id: int,
            under_water_trg: PublicKey,
    ):
        ix = dixs.cancel_order(
            user=self.keypair.public_key,
            trader_risk_group=under_water_trg,
            market_product_group=sdk.market_product_group,
            product=product.key,
            aaob_program=sdk.aaob_program,
            orderbook=product.orderbook,
            market_signer=product.market_signer,
            event_queue=product.event_queue,
            bids=product.bids,
            asks=product.asks,
            risk_engine_program=sdk.risk_engine_program,
            risk_model_configuration_acct=sdk.risk_model_configuration_acct,
            risk_output_register=sdk.risk_output_register,
            trader_risk_state_acct=self.trader_risk_state_acct,
            risk_signer=sdk.risk_signer,
            params=CancelOrderParams(order_id=order_id),
            system_program=SYS_PROGRAM_ID,
            remaining_accounts=None,
            program_id=self.dex_program,
        )

        return ix

    def replace(
            self,
            sdk: "SDKContext",
            product: Union[SDKProduct, PublicKey],
            order_id: int,
            side: aaob_state.Side,
            size: Union[dtys.Fractional, float],
            price: Union[dtys.Fractional, float],
            self_trade_behavior: aaob_state.SelfTradeBehavior = aaob_state.SelfTradeBehavior.DECREMENT_TAKE,
            order_type: dtys.OrderType = dtys.OrderType.LIMIT
    ) -> DexOrderSummary:
        cancel_ix = self._cancel_ix(sdk, product, order_id, self.account)
        place_ix = self._place_order_ix(sdk, product, side, size, price, self_trade_behavior, order_type,
                                        sdk.additional_risk_accts)

        trans_details = sdk.send_instructions(cancel_ix, place_ix, raise_on_error=False)
        if trans_details.error:
            raise Exception(trans_details.error_from_log)
        else:
            raw_summary = trans_details.emitted_logs["new-order:order-summary"]
            return SDKOrderSummary.from_dex_order_summary(sdk, DexOrderSummary.from_bytes(raw_summary),
                                                          product, side, size, price)

    def cancel_all_orders(
            self,
            sdk: "SDKContext",
            product_indices: List[int]
    ):
        trader_risk_group = self.get_trader_risk_group()
        print("Cancelling all orders")
        for n in product_indices:
            order_ids = []
            ptr = trader_risk_group.open_orders.products[n].head_index
            order = trader_risk_group.open_orders.orders[ptr]
            assert order.prev == SENTINEL
            while ptr != SENTINEL:
                order = trader_risk_group.open_orders.orders[ptr]
                assert order.id != 0
                order_ids.append(order.id)
                ptr = order.next

            if order_ids:
                print(f"Cancelling orders with ids {', '.join([str(o_id) for o_id in order_ids])}")
                # TODO parallelize?
                for order_id in order_ids:
                    self.cancel(sdk, sdk.products[n], order_id)
                    print(f"Cancelled order with id {order_id}")


@dataclass
class SDKContext:
    product_group_name: str  # max 16 chars
    trader_risk_state_account_len: int
    decimals: int
    # cached products, reload if necessary
    products: List[SDKProduct]
    # program_ids
    dex_program: PublicKey
    aaob_program: PublicKey
    risk_engine_program: PublicKey
    instruments_program: PublicKey
    # dummy_oracle_program_id: PublicKey
    fee_model_program: PublicKey
    # accts
    market_product_group: PublicKey
    capital_limits: PublicKey
    payer: Keypair
    market_product_group_vault: PublicKey
    vault_mint: PublicKey
    fee_model_configuration_acct: PublicKey
    risk_model_configuration_acct: PublicKey
    risk_signer: PublicKey
    fee_signer: PublicKey
    risk_output_register: PublicKey
    fee_output_register: PublicKey
    fee_collector: PublicKey
    additional_risk_accts: List[PublicKey]
    tx_opts: TxOpts = None

    @staticmethod
    def connect(client: Client,
                payer: Keypair,
                market_product_group_key: PublicKey,
                trader_risk_state_account_len: int = 0,
                dex_program_id: PublicKey = pids.DEX_PROGRAM_ID,
                aaob_program_id: PublicKey = pids.AOB_PROGRAM_ID,
                risk_engine_program_id: PublicKey = pids.RISK_ENGINE_PROGRAM_ID,
                instruments_program_id: PublicKey = pids.INSTRUMENTS_PROGRAM_ID,
                fee_model_program_id: PublicKey = pids.CONSTANT_FEES_MODEL_PROGRAM_ID,
                raise_on_error: bool = False,
                tx_opts: TxOpts = None,
                **kwargs):
        parser = AccountParser()
        parser.register_parser_from_account_enum(pids.DEX_PROGRAM_ID, DexAccounts)
        parser.register_parser(pids.AOB_PROGRAM_ID, aaob_state.account_parser)
        parser.register_parser_from_account_enum(pids.INSTRUMENTS_PROGRAM_ID, InstrumentAccounts)
        Context.init_globals(
            fee_payer=payer,
            client=client,
            signers=[(payer, "payer")],
            parser=parser,
            raise_on_error=raise_on_error,
        )

        mpg: MarketProductGroup = solana_utils.explore(market_product_group_key).data_obj
        print(f"Connecting to mpg {bytes(mpg.name).strip()}")

        capital_limits_key, _ = PublicKey.find_program_address(
            [b"capital_limits_state", bytes(market_product_group_key)],
            dex_program_id
        )

        s_account, _ = PublicKey.find_program_address(
            [b"s", bytes(market_product_group_key)],
            risk_engine_program_id
        )

        r_account, _ = PublicKey.find_program_address(
            [b"r", bytes(market_product_group_key)],
            risk_engine_program_id
        )

        sdk_context = SDKContext(
            product_group_name=bytes(mpg.name).decode("utf-8").strip(),
            trader_risk_state_account_len=trader_risk_state_account_len,
            decimals=mpg.decimals,
            # cached products reload if necessary
            products=[],
            # program_ids
            dex_program=dex_program_id,
            aaob_program=aaob_program_id,
            risk_engine_program=risk_engine_program_id,
            instruments_program=instruments_program_id,
            # dummy_oracle_program_id=None,
            fee_model_program=fee_model_program_id,
            # accts
            market_product_group=market_product_group_key,
            capital_limits=capital_limits_key,
            payer=payer,
            market_product_group_vault=daddrs.get_market_product_group_vault(market_product_group_key),
            vault_mint=mpg.vault_mint,
            fee_model_configuration_acct=mpg.fee_model_configuration_acct,
            risk_model_configuration_acct=mpg.risk_model_configuration_acct,
            risk_signer=daddrs.get_risk_signer(market_product_group_key),
            fee_signer=daddrs.get_risk_signer(market_product_group_key),
            risk_output_register=mpg.risk_output_register,
            fee_output_register=mpg.fee_output_register,
            fee_collector=mpg.fee_collector,
            additional_risk_accts=[s_account, r_account],
            tx_opts=tx_opts,
        )
        sdk_context.load_products()
        return sdk_context

    def list_trader_risk_groups(self) -> List[PublicKey]:
        account_discriminator_filter = types.MemcmpOpts(
            offset=0,
            bytes=str(base58.b58encode(
                int(DexAccounts.TRADER_RISK_GROUP).to_bytes(8, "little")
            ), 'utf-8')
        )
        mpg_filter = types.MemcmpOpts(
            offset=16,
            bytes=str(self.market_product_group.to_base58(), 'utf-8')
        )
        trader_filter = types.MemcmpOpts(
            offset=48,
            bytes=str(self.payer.public_key.to_base58(), 'utf-8')
        )
        response = Context.get_global_client().get_program_accounts(
            pubkey=self.dex_program,
            commitment=Confirmed,
            encoding="base64",
            data_slice=types.DataSliceOpts(offset=0, length=0),  # we don't need any data
            memcmp_opts=[account_discriminator_filter, mpg_filter, trader_filter]
        )
        trgs = []
        if "result" in response:
            for account in response["result"]:
                trgs.append(PublicKey(account["pubkey"]))
        return trgs

    def load_mpg(self) -> MarketProductGroup:
        return solana_utils.fetch_account_details(self.market_product_group).data_obj

    def load_products(self):
        mpg = self.load_mpg()
        products = []
        for prod in mpg.active_products():
            metadata = prod.metadata()
            orderbook: aaob_state.MarketState = fetch_account_details(metadata.orderbook).data_obj
            sdk_product = SDKProduct(
                metadata.product_key,
                bytes(metadata.name).decode('utf-8').strip(),
                orderbook=metadata.orderbook,
                asks=orderbook.asks,
                bids=orderbook.bids,
                event_queue=orderbook.event_queue,
                market_signer=daddrs.get_market_signer(metadata.product_key),
                stale_product=prod,
            )
            products.append(sdk_product)
        self.products = products

    def send_instructions(self, *ixs: TransactionInstruction, **kwargs):
        if self.tx_opts is not None:
            return solana_utils.send_instructions(*ixs, opts=self.tx_opts, **kwargs)
        else:
            return solana_utils.send_instructions(*ixs, **kwargs)

    def register_trader(self, keypair: Keypair, wallet: PublicKey):
        from solana.system_program import SYS_PROGRAM_ID
        trader_risk_group = Keypair.generate()
        trader_risk_state_acct = Keypair.generate()
        _ident = keypair.public_key.to_base58()[:8]
        Context.add_signers(
            (trader_risk_state_acct, f"{_ident}'s trader_risk_state_acct"),
            (trader_risk_group, f"{_ident}'s trader_risk_group)"),
        )
        trader_fee_state_acct = daddrs.get_trader_fee_state_acct(
            trader_risk_group.public_key,
            self.market_product_group,
            self.fee_model_program)

        fee_ix = fee_ixs.initialize_trader_acct_ix(
            program_id=self.fee_model_program,
            payer=self.payer.public_key,
            fee_model_config_acct=self.fee_model_configuration_acct,
            trader_fee_acct=trader_fee_state_acct,
            market_product_group=self.market_product_group,
            trader_risk_group=trader_risk_group.public_key,
            system_program=SYS_PROGRAM_ID)
        size = dtys.TraderRiskGroup.calc_max_size() + 8
        allocate_trg = system_program.create_account(
            system_program.CreateAccountParams(
                from_pubkey=self.payer.public_key,
                new_account_pubkey=trader_risk_group.public_key,
                lamports=solana_utils.calc_rent(size),
                space=size,
                program_id=self.dex_program,
            )
        )
        trg_init_ix = dixs.initialize_trader_risk_group(
            owner=keypair.public_key,
            trader_risk_group=trader_risk_group.public_key,
            trader_risk_state_acct=trader_risk_state_acct.public_key,
            trader_fee_state_acct=trader_fee_state_acct,
            market_product_group=self.market_product_group,
            risk_signer=self.risk_signer,
            risk_engine_program=self.risk_engine_program,
            program_id=self.dex_program,
            # **vars(self),
        )
        self.send_instructions(fee_ix, allocate_trg, trg_init_ix)
        return SDKTrader.connect(
            self,
            trader_risk_group.public_key,
            keypair,
            wallet,
            trader_risk_state_acct.public_key,
            dex_program=self.dex_program,
        )



import os

from dexteritysdk.local_env import *

import asyncio
import json
import base64
import argparse
import websockets as ws
import time

# dex
from dexteritysdk.dex.state import *
from dexteritysdk.dex.state import account_parser as dex_parser
import dexteritysdk.dex.actions as dex_actions

# derivatives
from dexteritysdk.instruments.state import *
from dexteritysdk.instruments.state import account_parser as instr_parser
import dexteritysdk.instruments.actions as instr_actions
from dexteritysdk.dummy_oracle.state import *
from dexteritysdk.utils.aob.state import *
from dexteritysdk.utils.aob.state import account_parser as aob_parser
from dexteritysdk.utils.solana import (
    explore,
    AccountParser,
    set_global_parser,
    get_global_client,
    set_global_client,
    set_signers,
)
import dexteritysdk.program_ids as pids

# RPC
from solana.rpc.commitment import *
from solana.rpc.api import Client
from solana.keypair import Keypair
from solana.publickey import PublicKey
from solana.sysvar import SYSVAR_CLOCK_PUBKEY


def load_account_as(pubkey: PublicKey, cls):
    client = get_global_client()
    info = client.get_account_info(pubkey, commitment=Confirmed)
    if info["result"]["value"] is None:
        return None
    data = base64.b64decode(info["result"]["value"]["data"][0])
    if len(data) == cls.calc_size() + 8:
        return cls.from_bytes(data[8:])
    return cls.from_bytes(data)


def get_account_info(event):
    try:
        return [event.event_data.callback_info.user_account]
    except:
        return [
            event.event_data.maker_callback_info.user_account,
            event.event_data.taker_callback_info.user_account,
        ]


curr_directory = os.path.dirname(os.path.abspath(__file__))
cfg_path = os.path.join(curr_directory, "config.json")
if os.path.exists(cfg_path):
    with open(cfg_path) as f:
        cfg = json.load(f)
        mpg_pubkey = cfg["mpg_key"]
else:
    ap = argparse.ArgumentParser()
    ap.add_argument("mpg_pubkey")
    args = ap.parse_args()
    mpg_pubkey = args.mpg_pubkey


payer = Keypair(b"nima" * 8)
market_product_group_pubkey = PublicKey(mpg_pubkey)
reward_target = payer.public_key
set_signers(payer)

WS_URI = "wss://api.devnet.solana.com/"
RPC_URL = "https://api.devnet.solana.com"

parser = AccountParser()
parser.register_parser(pids.DEX_PROGRAM_ID, dex_parser)
parser.register_parser(pids.INSTRUMENTS_PROGRAM_ID, instr_parser)
parser.register_parser(pids.AOB_PROGRAM_ID, aob_parser)
set_global_parser(parser)
set_global_client(Client(RPC_URL))


seq = 0


def get_rpc_payload(key):
    global seq
    payload = {
        "jsonrpc": "2.0",
        "id": seq,
        "method": "accountSubscribe",
        "params": [
            key,
            {"encoding": "base64", "commitment": "confirmed"},
        ],
    }
    seq += 1
    return payload


async def settle_funds(market_product_group):
    clock = load_account_as(SYSVAR_CLOCK_PUBKEY, Clock)
    for product in market_product_group.market_products:
        if product:
            product_pubkey = product.product_key
            derivative_metadata = load_account_as(product_pubkey, DerivativeMetadata)
            if not derivative_metadata:
                continue
            print(f"Considering Product {product_pubkey}")
            print(f"    Current Funding per Share: {product.cum_funding_per_share}")
            last_funding_ts = derivative_metadata.last_funding_time
            next_funding_ts = (
                last_funding_ts + derivative_metadata.minimum_funding_period
            )
            curr_ts = clock.unix_timestamp
            print(f"    Current Time: {curr_ts}")
            print(f"    Next Funding: {next_funding_ts} (last={last_funding_ts})")
            print(f"    Time Until Next Funding: {next_funding_ts - curr_ts}")
            if curr_ts < next_funding_ts:
                print("    -> Funding period has not elapsed")
                continue
            print("Oracle:", derivative_metadata.price_oracle)
            _, resp = instr_actions.settle_derivative(
                market_product_group_pubkey,
                product.product_key,
                payer.public_key,
                derivative_metadata.price_oracle,
            )
            print(resp.log_messages)
        else:
            print("all products settled!")
            break


async def run():
    while True:
        try:
            market_product_group = load_account_as(
                market_product_group_pubkey, MarketProductGroup
            )
            await settle_funds(market_product_group)
            time.sleep(1)
        except:
            continue


asyncio.run(run())



from typing import List
from dexteritysdk.program_ids import DEX_PROGRAM_ID, INSTRUMENTS_PROGRAM_ID

from solana.keypair import Keypair
from solana.publickey import PublicKey
from solana.transaction import Transaction

import dexteritysdk.instruments.instructions as ixs

from dexteritysdk.codegen.dex.types import Fractional
from dexteritysdk.codegen.instruments import instructions as iixs
from dexteritysdk.codegen.instruments import types as its
from dexteritysdk.utils.solana import (
    actionify,
    Context,
)
from dexteritysdk.codegen.instruments.types import (
    InstrumentType,
    OracleType,
)


def _calc_rent(space, client=None):
    if client is None:
        client = Context.get_global_client()
    return client.get_minimum_balance_for_rent_exemption(space)["result"]


def extract_acct_addr(resp, idx=0):
    addr = resp.instructions[0]["accounts"][idx]
    exists = False
    if resp.error:
        error_ix, error_info = resp.error["InstructionError"]
        if error_ix == 0 and error_info["Custom"] == 0:
            exists = True
    else:
        exists = True

    if exists:
        return addr, resp
    else:
        return None, resp


@actionify(post_process=lambda x: extract_acct_addr(x, idx=0))
def initialize_derivative(
    price_oracle: PublicKey,
    market_product_group: PublicKey,
    payer: PublicKey,
    instrument_type: InstrumentType,
    strike: float,
    full_funding_period: int,
    minimum_funding_period: int,
    initialization_time: int,
    oracle_type: OracleType,
    close_authority: PublicKey,
    clock: PublicKey = None,
    **kwargs,
):
    params = its.InitializeDerivativeParams(
        instrument_type=instrument_type,
        strike=Fractional.to_decimal(strike),
        full_funding_period=full_funding_period,
        minimum_funding_period=minimum_funding_period,
        initialization_time=initialization_time,
        close_authority=close_authority,
        oracle_type=oracle_type,
    )
    return Transaction().add(
        iixs.initialize_derivative(
            derivative_metadata=ixs.get_derivative_key(
                price_oracle=price_oracle,
                market_product_group=market_product_group,
                instrument_type=instrument_type,
                strike=strike,
                full_funding_period=full_funding_period,
                minimum_funding_period=minimum_funding_period,
                initialization_time=initialization_time,
            )[0],
            price_oracle=price_oracle,
            market_product_group=market_product_group,
            payer=payer,
            clock=clock,
            params=params,
        ),
    )


@actionify(post_process=lambda x: extract_acct_addr(x, idx=0))
def initialize_fixed_income(
    face_value: int,
    market_product_group: PublicKey,
    payer: Keypair,
    coupon_dates: List[int],
    coupon_rate: int,
    maturity_date: int,
    initialization_time: int,
    close_authority: PublicKey,
):
    return iixs.initialize_fixed_income(
        fixed_income_metadata=ixs.get_fixed_income_key(
            market_product_group=market_product_group,
            initialization_time=initialization_time,
            coupon_rate=coupon_rate,
            maturity_date=maturity_date
        )[0],
        market_product_group=market_product_group,
        payer=payer.public_key,
        params=its.InitializeFixedIncomeParams(
            face_value=face_value,
            coupon_rate=coupon_rate,
            initialization_time=initialization_time,
            coupon_dates=coupon_dates,
            maturity_date=maturity_date,
            close_authority=close_authority,
        )
    )


_tick = 0
@actionify
def settle_derivative(
    market_product_group: PublicKey,
    derivative_metadata: PublicKey,
    payer: PublicKey,
    price_oracle: PublicKey,
    clock: PublicKey = None,
):
    _tick += 1
    return Transaction(fee_payer=payer).add(
        iixs.settle_derivative(
            market_product_group,
            derivative_metadata=derivative_metadata,
            price_oracle=price_oracle,
            clock=clock,
            dex_program=DEX_PROGRAM_ID,
            tick=_tick,
        ),
    )


@actionify(post_process=lambda x: extract_acct_addr(x, idx=1))
def settle_fixed_income_ix(
    market_product_group: PublicKey,
    fixed_income_metadata: PublicKey,
):
    return iixs.settle_fixed_income(
        market_product_group=market_product_group,
        fixed_income_metadata=fixed_income_metadata,
        dex_program=DEX_PROGRAM_ID,
    )
